#
# AstroImage.py -- Abstraction of an astronomical data image.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import sys
import traceback
import warnings
from collections import OrderedDict

import numpy as np

from ginga.util import wcs, wcsmod
from ginga.BaseImage import BaseImage, ImageError, Header
from ginga.misc import Bunch


class AstroHeader(Header):
    pass


class AstroImage(BaseImage):
    """
    Abstraction of an astronomical data (image).

    NOTE: this module is NOT thread-safe!
    """
    # class variables for WCS and IO can be set
    wcsClass = None
    ioClass = None

    @classmethod
    def set_wcsClass(cls, klass):
        cls.wcsClass = klass

    @classmethod
    def set_ioClass(cls, klass):
        cls.ioClass = klass

    def __init__(self, data_np=None, metadata=None, logger=None,
                 name=None, wcsclass=None, ioclass=None,
                 inherit_primary_header=False, save_primary_header=True):

        BaseImage.__init__(self, data_np=data_np, metadata=metadata,
                           logger=logger, name=name)

        # wcsclass specifies a pluggable WCS module
        if wcsclass is None:
            if self.wcsClass is None:
                wcsclass = wcsmod.WCS
            else:
                wcsclass = self.wcsClass
        self.wcs = wcsclass(self.logger)

        # ioclass specifies a pluggable IO module
        if ioclass is None:
            if self.ioClass is None:
                from ginga.util import io_fits
                ioclass = io_fits.fitsLoaderClass
            else:
                ioclass = self.ioClass

        self.io = ioclass(self.logger)

        self.inherit_primary_header = inherit_primary_header
        self.save_primary_header = inherit_primary_header or save_primary_header
        if self.save_primary_header:
            # User wants to save/inherit from primary header--this will hold it
            self._primary_hdr = AstroHeader()
        else:
            self._primary_hdr = None

        if metadata is not None:
            header = self.get_header()
            self.wcs.load_header(header)

        # For navigating multidimensional data
        self.naxispath = []
        self.revnaxis = []
        self._md_data = None

    def setup_data(self, data, naxispath=None):
        # initialize data attribute to something reasonable
        if data is None:
            data = np.zeros((0, 0))
        elif not isinstance(data, np.ndarray):
            data = np.zeros((0, 0))
        elif 0 in data.shape:
            data = np.zeros((0, 0))

        # this is a handle to the full data array
        self._md_data = data

        # this will get reset in set_naxispath() if array is
        # multidimensional
        self._data = self._md_data

        if naxispath is None:
            naxispath = []

        # Set naxispath to drill down to first 2D data slice
        if len(naxispath) == 0:
            naxispath = ([0] * (len(data.shape) - 2))

        self.set_naxispath(naxispath)

    def load_hdu(self, hdu, fobj=None, naxispath=None,
                 inherit_primary_header=None):

        self.clear_metadata()

        # collect HDU header
        ahdr = self.get_header()
        self._copy_hdu_header(hdu, ahdr)

        # Set PRIMARY header
        if inherit_primary_header is None:
            inherit_primary_header = self.inherit_primary_header
        else:  # This ensures get_header() is consistent
            self.inherit_primary_header = inherit_primary_header

        save_primary_header = (self.save_primary_header or
                               inherit_primary_header)
        if save_primary_header and (fobj is not None):
            if self._primary_hdr is None:
                self._primary_hdr = AstroHeader()

            self._copy_hdu_header(fobj[0], self._primary_hdr)

        self.setup_data(hdu.data, naxispath=naxispath)

        # Try to make a wcs object on the header
        if hasattr(self, 'wcs') and self.wcs is not None:
            self.wcs.load_header(hdu.header, fobj=fobj)

    def load_nddata(self, ndd, naxispath=None):
        """Load from an astropy.nddata.NDData object.
        """
        self.clear_metadata()

        # Make a header based on any NDData metadata
        ahdr = self.get_header()
        ahdr.update(ndd.meta)

        self.setup_data(ndd.data, naxispath=naxispath)

        if ndd.wcs is None:
            # no wcs in ndd obj--let's try to make one from the header
            self.wcs = wcsmod.WCS(logger=self.logger)
            self.wcs.load_header(ahdr)
        else:
            # already have a valid wcs in the ndd object
            # we assume it needs an astropy compatible wcs
            wcsinfo = wcsmod.get_wcs_class('astropy')
            self.wcs = wcsinfo.wrapper_class(logger=self.logger)
            self.wcs.load_nddata(ndd)

    def load_file(self, filespec, **kwargs):

        if self.io is None:
            raise ImageError("No IO loader defined")

        self.io.load_file(filespec, dstobj=self, **kwargs)

    def load_data(self, data_np, naxispath=None, metadata=None):

        self.clear_metadata()

        self.setup_data(data_np, naxispath=naxispath)

        if metadata is not None:
            self.update_metadata(metadata)

    def load_buffer(self, buf, dims, dtype, byteswap=False,
                    naxispath=None, metadata=None):
        data = np.fromstring(buf, dtype=dtype)
        if byteswap:
            data.byteswap(True)
        data = data.reshape(dims)

        self.load_data(data, naxispath=naxispath, metadata=metadata)

    def get_mddata(self):
        return self._md_data

    def set_naxispath(self, naxispath):
        """Choose a slice out of multidimensional data.
        """
        revnaxis = list(naxispath)
        revnaxis.reverse()

        # construct slice view and extract it
        ndim = min(self.ndim, 2)
        view = tuple(revnaxis + [slice(None)] * ndim)
        data = self.get_mddata()[view]

        if len(data.shape) not in (1, 2):
            raise ImageError(
                "naxispath does not lead to a 1D or 2D slice: {}".format(naxispath))

        self.naxispath = naxispath
        self.revnaxis = revnaxis

        self.set_data(data)

    def set_wcs(self, wcs):
        self.wcs = wcs

    def set_io(self, io):
        self.io = io

    def get_data_size(self):
        return self.get_size()

    def get_header(self, create=True, include_primary_header=None):
        try:
            # By convention, the fits header is stored in a dictionary
            # under the metadata keyword 'header'
            hdr = self.metadata['header']

            if include_primary_header is None:
                include_primary_header = self.inherit_primary_header

            if include_primary_header and self._primary_hdr is not None:
                # Inherit PRIMARY header for display but keep metadata intact
                displayhdr = AstroHeader()
                for key in hdr.keyorder:
                    card = hdr.get_card(key)
                    bnch = displayhdr.__setitem__(card.key, card.value)
                    bnch.comment = card.comment
                for key in self._primary_hdr.keyorder:
                    if key not in hdr:
                        card = self._primary_hdr.get_card(key)
                        bnch = displayhdr.__setitem__(card.key, card.value)
                        bnch.comment = card.comment
            else:
                # Normal, separate header
                displayhdr = hdr

        except KeyError as e:
            if not create:
                raise e
            hdr = AstroHeader()
            self.metadata['header'] = hdr
            displayhdr = hdr

        return displayhdr

    def get_keyword(self, kwd, *args):
        """Get an item from the fits header, if any."""
        try:
            kwds = self.get_header()
            return kwds[kwd]
        except KeyError:
            # return a default if there is one
            if len(args) > 0:
                return args[0]
            raise KeyError(kwd)

    def get_keywords_list(self, *args):
        return list(map(self.get_keyword, args))

    def set_keyword(self, kwd, value, create=True):
        kwds = self.get_header(create=create)
        kwd = kwd.upper()
        if not create:
            prev = kwds[kwd]  # noqa, this raises KeyError
        kwds[kwd] = value

    def update_keywords(self, keyDict):
        hdr = self.get_header()
        # Upcase all keywords
        for kwd, val in keyDict.items():
            hdr[kwd.upper()] = val

        # Try to make a wcs object on the header
        if hasattr(self, 'wcs') and self.wcs is not None:
            self.wcs.load_header(hdr)

    def set_keywords(self, **kwds):
        """Set an item in the fits header, if any."""
        return self.update_keywords(kwds)

    def update_data(self, data_np, metadata=None, astype=None):
        """DO NOT USE: this method will be deprecated!
        """
        self.set_data(data_np.copy(), metadata=metadata,
                      astype=astype)

    def update_metadata(self, key_dict):
        for key, val in key_dict.items():
            self.metadata[key] = val

        # refresh the WCS
        if hasattr(self, 'wcs') and self.wcs is not None:
            header = self.get_header()
            self.wcs.load_header(header)

    def has_primary_header(self):
        return self._primary_hdr is not None

    def clear_all(self):
        # clear metadata and data
        super(AstroImage, self).clear_all()

        # unreference full data array
        self._md_data = self._data

    def transfer(self, other, astype=None):
        data = self._get_data()
        other.update_data(data, astype=astype)
        other.update_metadata(self.metadata)

    def copy(self, astype=None):
        data = self._get_data()
        other = AstroImage(data, logger=self.logger)
        self.transfer(other, astype=astype)
        return other

    def as_nddata(self, nddata_class=None):
        "Return a version of ourself as an astropy.nddata.NDData object"
        if nddata_class is None:
            from astropy.nddata import NDData
            nddata_class = NDData

        # transfer header, preserving ordering
        ahdr = self.get_header()
        header = OrderedDict(ahdr.items())
        data = self.get_mddata()

        wcs = None
        if hasattr(self, 'wcs') and self.wcs is not None:
            # for now, assume self.wcs wraps an astropy wcs object
            wcs = self.wcs.wcs

        ndd = nddata_class(data, wcs=wcs, meta=header)
        return ndd

    def as_hdu(self):
        "Return a version of ourself as an astropy.io.fits.PrimaryHDU object"
        from astropy.io import fits

        # transfer header, preserving ordering
        ahdr = self.get_header()
        header = fits.Header(ahdr.items())
        data = self.get_mddata()

        hdu = fits.PrimaryHDU(data=data, header=header)
        return hdu

    def _copy_hdu_header(self, hdu, ahdr):
        """Copy a FITS header from an astropy.io.fits.PrimaryHDU object
        into a ginga.AstroImage.AstroHeader object.
        """
        header = hdu.header
        for card in header.cards:
            if len(card.keyword) == 0:
                continue
            ahdr.set_card(card.keyword, card.value, comment=card.comment)

    def astype(self, type_name):
        """Convert AstroImage object to some other kind of object.
        """
        if type_name == 'nddata':
            return self.as_nddata()

        if type_name == 'hdu':
            return self.as_hdu()

        raise ValueError("Unrecognized conversion type '%s'" % (type_name))

    def save_as_file(self, filepath, **kwdargs):
        data = self._get_data()
        header = self.get_header()
        self.io.save_as_file(filepath, data, header, **kwdargs)

    def pixtocoords(self, x, y, system=None, coords='data'):
        args = [x, y] + self.revnaxis
        return self.wcs.pixtocoords(args, system=system, coords=coords)

    def spectral_coord(self, coords='data'):
        args = [0, 0] + self.revnaxis
        return self.wcs.spectral_coord(args, coords=coords)

    def pixtoradec(self, x, y, format='deg', coords='data'):
        args = [x, y] + self.revnaxis
        ra_deg, dec_deg = self.wcs.pixtoradec(args, coords=coords)

        if format == 'deg':
            return ra_deg, dec_deg
        return wcs.deg2fmt(ra_deg, dec_deg, format)

    def radectopix(self, ra_deg, dec_deg, format='deg', coords='data'):
        if format != 'deg':
            # convert coordinates to degrees
            ra_deg = wcs.lon_to_deg(ra_deg)
            dec_deg = wcs.lat_to_deg(dec_deg)
        return self.wcs.radectopix(ra_deg, dec_deg, coords=coords,
                                   naxispath=self.revnaxis)

    # -----> TODO:
    #   This section has been merged into ginga.util.wcs or
    #   ginga.util.mosaic .  Deprecate it here.
    #
    def get_starsep_XY(self, x1, y1, x2, y2):
        warnings.warn("This function has been deprecated--"
                      "use the version in ginga.util.wcs",
                      PendingDeprecationWarning)
        return wcs.get_starsep_XY(self, x1, y1, x2, y2)

    def calc_radius_xy(self, x, y, radius_deg):
        warnings.warn("This function has been deprecated--"
                      "use the version in ginga.util.wcs",
                      PendingDeprecationWarning)
        return wcs.calc_radius_xy(self, x, y, radius_deg)

    def calc_radius_deg2pix(self, ra_deg, dec_deg, delta_deg,
                            equinox=None):
        warnings.warn("This function has been deprecated--"
                      "use the version in ginga.util.wcs",
                      PendingDeprecationWarning)
        return wcs.calc_radius_deg2pix(self, ra_deg, dec_deg, delta_deg,
                                       equinox=equinox)

    def add_offset_xy(self, x, y, delta_deg_x, delta_deg_y):
        warnings.warn("This function has been deprecated--"
                      "use the version in ginga.util.wcs",
                      PendingDeprecationWarning)
        return wcs.add_offset_xy(self, x, y, delta_deg_x, delta_deg_y)

    def calc_radius_center(self, delta_deg):
        warnings.warn("This function has been deprecated--"
                      "use the version in ginga.util.wcs",
                      PendingDeprecationWarning)
        return wcs.calc_radius_center(self, delta_deg)

    def calc_compass(self, x, y, len_deg_e, len_deg_n):
        warnings.warn("This function has been deprecated--"
                      "use the version in ginga.util.wcs",
                      PendingDeprecationWarning)
        return wcs.calc_compass(self, x, y, len_deg_e, len_deg_n)

    def calc_compass_radius(self, x, y, radius_px):
        warnings.warn("This function has been deprecated--"
                      "use the version in ginga.util.wcs",
                      PendingDeprecationWarning)
        return wcs.calc_compass_radius(self, x, y, radius_px)

    def calc_compass_center(self):
        warnings.warn("This function has been deprecated--"
                      "use the version in ginga.util.wcs",
                      PendingDeprecationWarning)
        return wcs.calc_compass_center(self)

    def get_wcs_rotation_deg(self):
        warnings.warn("This function has been deprecated--"
                      "use get_rotation_and_scale in ginga.util.wcs",
                      PendingDeprecationWarning)
        header = self.get_header()
        (rot, cdelt1, cdelt2) = wcs.get_rotation_and_scale(header)
        return rot

    def mosaic_inline(self, imagelist, bg_ref=None, trim_px=None,
                      merge=False, allow_expand=True, expand_pad_deg=0.01,
                      max_expand_pct=None,
                      update_minmax=True, suppress_callback=False):
        warnings.warn("This function has been deprecated--"
                      "use the version in ginga.util.mosaic",
                      PendingDeprecationWarning)
        from ginga.util import mosaic
        return mosaic.mosaic_inline(self, imagelist, bg_ref=bg_ref,
                                    trim_px=trim_px, merge=merge,
                                    allow_expand=allow_expand,
                                    expand_pad_deg=expand_pad_deg,
                                    max_expand_pct=max_expand_pct,
                                    update_minmax=update_minmax,
                                    suppress_callback=suppress_callback)
    #
    # <----- TODO: deprecate

    def info_xy(self, data_x, data_y, settings):
        # Get the value under the data coordinates
        try:
            # We report the value across the pixel, even though the coords
            # change halfway across the pixel
            value = self.get_data_xy(int(data_x + 0.5), int(data_y + 0.5))

        except Exception as e:
            value = None

        system = settings.get('wcs_coords', None)
        format = settings.get('wcs_display', 'sexagesimal')
        ra_lbl, dec_lbl = chr(945), chr(948)

        # Calculate WCS coords, if available
        try:
            if self.wcs is None:
                self.logger.debug("No WCS for this image")
                ra_txt = dec_txt = 'NO WCS'

            elif self.wcs.coordsys == 'raw':
                self.logger.debug("No coordinate system determined")
                ra_txt = dec_txt = 'NO WCS'

            elif self.wcs.coordsys == 'pixel':
                args = [data_x, data_y] + self.revnaxis
                x, y = self.wcs.pixtosystem(args, system=system, coords='data')
                ra_txt = "%+.3f" % (x)
                dec_txt = "%+.3f" % (y)
                ra_lbl, dec_lbl = "X", "Y"

            else:
                args = [data_x, data_y] + self.revnaxis

                lon_deg, lat_deg = self.wcs.pixtosystem(
                    args, system=system, coords='data')

                if format == 'sexagesimal':
                    if system in ('galactic', 'ecliptic'):
                        sign, deg, min, sec = wcs.degToDms(lon_deg,
                                                           isLatitude=False)
                        ra_txt = '+%03d:%02d:%06.3f' % (deg, min, sec)
                    else:
                        deg, min, sec = wcs.degToHms(lon_deg)
                        ra_txt = '%02d:%02d:%06.3f' % (deg, min, sec)

                    sign, deg, min, sec = wcs.degToDms(lat_deg)
                    if sign < 0:
                        sign = '-'
                    else:
                        sign = '+'
                    dec_txt = '%s%02d:%02d:%06.3f' % (sign, deg, min, sec)

                else:
                    ra_txt = '%+10.7f' % (lon_deg)
                    dec_txt = '%+10.7f' % (lat_deg)

                if system == 'galactic':
                    ra_lbl, dec_lbl = "l", "b"
                elif system == 'ecliptic':
                    ra_lbl, dec_lbl = chr(0x03BB), chr(0x03B2)
                elif system == 'helioprojective':
                    ra_txt = "%+5.3f" % (lon_deg * 3600)
                    dec_txt = "%+5.3f" % (lat_deg * 3600)
                    ra_lbl, dec_lbl = "x-Solar", "y-Solar"

        except Exception as e:
            self.logger.warning("Bad coordinate conversion: %s" % (
                str(e)))
            ra_txt = dec_txt = 'BAD WCS'
            try:
                # log traceback, if possible
                (type_, value_, tb) = sys.exc_info()
                tb_str = "".join(traceback.format_tb(tb))
                self.logger.error("Traceback:\n%s" % (tb_str))
            except Exception:
                tb_str = "Traceback information unavailable."
                self.logger.error(tb_str)

        info = Bunch.Bunch(itype='astro', data_x=data_x, data_y=data_y,
                           x=data_x, y=data_y,
                           ra_txt=ra_txt, dec_txt=dec_txt,
                           ra_lbl=ra_lbl, dec_lbl=dec_lbl,
                           value=value)
        return info

# END



import logging
import warnings

import astropy
import numpy as np
import pytest
from astropy.utils.introspection import minversion
from numpy.testing import assert_allclose

from ginga import AstroImage
from ginga.util import wcsmod

# TODO: Add a test for native GWCS object.

_logger = logging.getLogger("TestWCS")
_wcsmods = ('kapteyn', 'starlink', 'astlib', 'astropy', 'astropy_ape14')
_hdr = {'2d': {'ADC-END': 6.28,
               'ADC-STR': 6.16,
               'ADC-TYPE': 'IN',
               'AIRMASS': 1.0526,
               'ALTITUDE': 72.142,
               'AUTOGUID': 'ON',
               'AZIMUTH': 282.679,
               'BIN-FCT1': 1,
               'BIN-FCT2': 1,
               'BITPIX': -32,
               'BLANK': -32768,
               'BUNIT': 'ADU',
               'CD1_1': -5.611e-05,
               'CD1_2': 0.0,
               'CD2_1': 0.0,
               'CD2_2': 5.611e-05,
               'CDELT1': -5.611e-05,
               'CDELT2': 5.611e-05,
               'COADD': 1,
               'CRPIX1': 5276.0,
               'CRPIX2': 25.0,
               'CRVAL1': 299.91736667,
               'CRVAL2': 22.68769444,
               'CTYPE1': 'RA---TAN',
               'CTYPE2': 'DEC--TAN',
               'CUNIT1': 'degree',
               'CUNIT2': 'degree',
               'DATA-TYP': 'OBJECT',
               'DATASET': 'DS000',
               'DATE-OBS': '2009-08-22',
               'DEC': '+22:41:15.70',
               'DEC2000': '+22:41:15.70',
               'DET-A01': 90.0,
               'DET-ID': 6,
               'DET-P101': -79.14,
               'DET-P201': -0.375,
               'DET-TMAX': 0.0,
               'DET-TMED': 0.0,
               'DET-TMIN': 0.0,
               'DET-TMP': 172.74,
               'DET-VER': 'spcam20080721',
               'DETECTOR': 'chihiro',
               'DOM-HUM': 12.4,
               'DOM-PRS': 622.3,
               'DOM-TMP': 276.35,
               'DOM-WND': 0.6,
               'EFP-MIN1': 9,
               'EFP-MIN2': 49,
               'EFP-RNG1': 2256,
               'EFP-RNG2': 4177,
               'EQUINOX': 2000.0,
               'EXP-ID': 'SUPE01118760',
               'EXP1TIME': 90.0,
               'EXPTIME': 90.0,
               'EXTEND': False,
               'FILTER01': 'W-J-B',
               'FOC-POS': 'Prime',
               'FOC-VAL': 7.14,
               'FRAMEID': 'SUPA01118766',
               'GAIN': 3.73,
               'HST': '23:34:25.911',
               'HST-END': '23:35:55.010',
               'HST-STR': '23:34:25.911',
               'INR-END': -174.487,
               'INR-STR': -174.239,
               'INS-VER': 'Messia5/sup080721',
               'INST-PA': 90.0,
               'INSTRUME': 'SuprimeCam',
               'LONGPOLE': 180.0,
               'LST': '21:15:48.968',
               'LST-END': '21:17:18.311',
               'LST-STR': '21:15:48.968',
               'M2-ANG1': 1.5,
               'M2-ANG2': -0.0,
               'M2-ANG3': 0.0,
               'M2-POS1': -0.753,
               'M2-POS2': -2.1,
               'M2-POS3': 8.205,
               'MJD': 55065.398914,
               'MJD-END': 55065.399945,
               'MJD-STR': 55065.398914,
               'NAXIS': 2,
               'NAXIS1': 2272,
               'NAXIS2': 4273,
               'OBJECT': 'M27',
               'OBS-ALOC': 'Observation',
               'OBS-MOD': 'IMAG_N_VGW',
               'OBSERVAT': 'NAOJ',
               'OBSERVER': 'Jeschke, Inagaki, Streeper, Yagi, Nakata',
               'OUT-HUM': 13.1,
               'OUT-PRS': 622.3,
               'OUT-TMP': 275.95,
               'OUT-WND': 6.0,
               'PRD-MIN1': 1,
               'PRD-MIN2': 1,
               'PRD-RNG1': 2272,
               'PRD-RNG2': 4273,
               'PROP-ID': 'o99005',
               'RA': '19:59:40.168',
               'RA2000': '19:59:40.168',
               'RADECSYS': 'FK5',
               'SECZ-END': 1.053,
               'SECZ-STR': 1.051,
               'SEEING': 0.29,
               'SIMPLE': True,
               'S_AG-DEC': 'N/A',
               'S_AG-EQN': 2000.0,
               'S_AG-OBJ': 'N/A',
               'S_AG-R': 999.99,
               'S_AG-RA': 'N/A',
               'S_AG-TH': 999.99,
               'S_AG-X': 109.97,
               'S_AG-Y': 19.3,
               'S_BCTAVE': 999.999,
               'S_BCTSD': 999.999,
               'S_DELTAD': 0.0,
               'S_DELTAZ': 0.0,
               'S_EFMN11': 9,
               'S_EFMN12': 49,
               'S_EFMN21': 617,
               'S_EFMN22': 49,
               'S_EFMN31': 1145,
               'S_EFMN32': 49,
               'S_EFMN41': 1753,
               'S_EFMN42': 49,
               'S_EFMX11': 520,
               'S_EFMX12': 4225,
               'S_EFMX21': 1128,
               'S_EFMX22': 4225,
               'S_EFMX31': 1656,
               'S_EFMX32': 4225,
               'S_EFMX41': 2264,
               'S_EFMX42': 4225,
               'S_ETMAX': 0.0,
               'S_ETMED': 273.15,
               'S_ETMIN': 0.0,
               'S_FRMPOS': '0001',
               'S_GAIN1': 3.73,
               'S_GAIN2': 2.95,
               'S_GAIN3': 3.1,
               'S_GAIN4': 3.17,
               'S_M2OFF1': 0.0,
               'S_M2OFF2': 0.0,
               'S_M2OFF3': 7.14,
               'S_OSMN11': 521,
               'S_OSMN12': 1,
               'S_OSMN21': 569,
               'S_OSMN22': 1,
               'S_OSMN31': 1657,
               'S_OSMN32': 1,
               'S_OSMN41': 1705,
               'S_OSMN42': 1,
               'S_OSMX11': 568,
               'S_OSMX12': 48,
               'S_OSMX21': 616,
               'S_OSMX22': 48,
               'S_OSMX31': 1704,
               'S_OSMX32': 48,
               'S_OSMX41': 1752,
               'S_OSMX42': 48,
               'S_SENT': False,
               'S_UFNAME': 'object060_chihiro.fits',
               'S_XFLIP': False,
               'S_YFLIP': True,
               'TELESCOP': 'Subaru',
               'TELFOCUS': 'P_OPT',
               'TIMESYS': 'UTC',
               'UT': '09:34:25.911',
               'UT-END': '09:35:55.010',
               'UT-STR': '09:34:25.911',
               'WCS-ORIG': 'SUBARU Toolkit',
               'WEATHER': 'Fine',
               'ZD-END': 18.2,
               'ZD-STR': 17.858},
        '3d': {'SIMPLE': True,
               'BITPIX': 16,
               'NAXIS': 3,
               'NAXIS1': 100,
               'NAXIS2': 100,
               'NAXIS3': 101,
               'BLOCKED': True,
               'CDELT1': -7.165998823E-03,
               'CRPIX1': 5.1E+01,
               'CRVAL1': -5.12820847959E+01,
               'CTYPE1': 'RA---NCP',
               'CUNIT1': 'deg',
               'CDELT2': 7.165998823E-03,
               'CRPIX2': 5.1E+01,
               'CRVAL2': 6.01538880206E+01,
               'CTYPE2': 'DEC--NCP',
               'CUNIT2': 'deg',
               'CDELT3': 4.199999809,
               'CRPIX3': -2.0E+01,
               'CRVAL3': -2.43E+02,
               'CTYPE3': 'VOPT',
               'CUNIT3': 'km/s',
               'EPOCH': 2.0E+03,
               'FREQ0': 1.420405758370E+09,
               'BUNIT': 'JY/BEAM ',
               'BMAJ': 1.82215739042E-02,
               'BMIN': 1.76625289023E-02,
               'BTYPE': 'intensity',
               'BPA': -7.41641769409E+01,
               'NITERS': 2626643,
               'LWIDTH': 4.19999980927E+00,
               'LSTEP': 4.19999980927E+00,
               'LSTART': -2.43E+02,
               'VOBS': -1.95447368244E+00,
               'LTYPE': 'velocity',
               'SPECSYS': 'BARYCENT'}}
img_dict = {}


def setup_module():
    """Create objects once and re-use throughout this module."""
    global img_dict

    if minversion(astropy, '3.1'):
        USE_APE14 = True
    else:
        USE_APE14 = False

    img_dict = {}
    for modname in _wcsmods:
        if modname == 'astropy_ape14' and not USE_APE14:
            continue
        if not wcsmod.use(modname, raise_err=False):
            continue
        img_dict[modname] = {}
        for dim in _hdr.keys():
            w = wcsmod.WCS(_logger)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                w.load_header(_hdr[dim])
            img = AstroImage.AstroImage(logger=_logger)
            img.wcs = w
            if dim == '2d':
                img.revnaxis = []
                img.naxispath = []
            else:  # 3d
                img.revnaxis = [0]
                img.naxispath = [0]
            img_dict[modname][dim] = img


@pytest.mark.parametrize('modname', _wcsmods)
def test_scalar_2d(modname):
    if modname not in img_dict:
        pytest.skip("WCS '{}' not available".format(modname))

    img = img_dict[modname]['2d']

    xy_v1 = (120, 100)
    radec_deg_v1 = (300.2308791294835, 22.691653517073615)

    # If this works here, should already work for other cases.
    assert img.wcs.has_valid_wcs()

    # 0.01% agreement is good enough across different libraries.
    radec = img.pixtoradec(*xy_v1)
    assert_allclose(radec, radec_deg_v1, rtol=1e-4)

    xy = img.radectopix(*radec_deg_v1)
    if modname == 'astropy_ape14':
        # TODO: Remove rtol when load_header is fixed.
        assert_allclose(xy, xy_v1, rtol=0.01)
    else:
        assert_allclose(xy, xy_v1)

    gal = img.wcs.pixtosystem(xy_v1, system='galactic')
    assert_allclose(gal, (60.97030081935234, -3.9706229385605307), rtol=1e-4)


@pytest.mark.parametrize('modname', _wcsmods)
def test_vectorized_2d(modname):
    if modname not in img_dict:
        pytest.skip("WCS '{}' not available".format(modname))

    img = img_dict[modname]['2d']

    xy_v1 = [(0, 0), (120, 100)]
    radec_deg_v1 = np.array([(300.2381639, 22.68602823),
                             (300.2308791294835, 22.691653517073615)])
    gal_v1 = np.array([(60.96903325, -3.97929572),
                       (60.97030081935234, -3.9706229385605307)])

    # 0.01% agreement is good enough across different libraries.
    radec = img.wcs.datapt_to_wcspt(xy_v1)
    assert_allclose(radec, radec_deg_v1, rtol=1e-4)

    xy = img.wcs.wcspt_to_datapt(radec_deg_v1)
    assert_allclose(xy, xy_v1, atol=7e-5)

    if modname == 'astlib':
        with pytest.raises(NotImplementedError):
            img.wcs.datapt_to_system(xy_v1, system='galactic')
    else:
        gal = img.wcs.datapt_to_system(xy_v1, system='galactic')
        if modname in ('astropy', 'astropy_ape14'):
            assert_allclose(gal.l.degree, gal_v1[:, 0])
            assert_allclose(gal.b.degree, gal_v1[:, 1])
        else:
            assert_allclose(gal, gal_v1, rtol=1e-4)


@pytest.mark.parametrize('modname', _wcsmods)
def test_scalar_3d(modname):
    if modname not in img_dict:
        pytest.skip("WCS '{}' not available".format(modname))

    img = img_dict[modname]['3d']
    idxs = (0, 0, 0)
    ra_deg_v1 = -50.569931842112965
    dec_deg_v1 = 59.79236398619401
    vel_v1 = -159000.00382

    if modname == 'astlib':
        with pytest.raises(wcsmod.common.WCSError):
            img.spectral_coord(idxs)
    else:
        c = img.spectral_coord(idxs)
        if modname == 'starlink':
            assert_allclose(c, vel_v1 * 1e-3)
        elif modname == 'astropy_ape14':
            # TODO: Remove rtol with load_header() is fixed.
            assert_allclose(c, vel_v1, rtol=0.03)
        else:
            assert_allclose(c, vel_v1)

    # 0.01% agreement is good enough across different libraries.
    # RA can be off by 360 degrees and still be valid.
    c = img.pixtoradec(*idxs[:2])
    assert (np.allclose(c, (ra_deg_v1, dec_deg_v1), rtol=1e-4) or
            np.allclose(c, (ra_deg_v1 + 360, dec_deg_v1), rtol=1e-4))

    px = img.radectopix(*c)
    assert_allclose(px, idxs[:2], atol=1e-3)

    c = img.wcs.pixtosystem(idxs, system='galactic')
    assert_allclose(c, (95.62934261967311, 11.172927294480449), rtol=1e-4)


@pytest.mark.parametrize('modname', _wcsmods)
def test_vectorized_3d(modname):
    if modname not in img_dict:
        pytest.skip("WCS '{}' not available".format(modname))

    img = img_dict[modname]['3d']

    xy_v1 = [(0, 0), (120, 100)]
    nxp = [0]
    radec_deg_v1 = np.array([[-50.5699318, 59.7923640],
                             [-52.3010162, 60.5064254]])
    gal_v1 = np.array([(95.62934262, 11.17292729),
                       (95.72174081, 12.28825976)])

    if modname == 'kapteyn':
        vel_v1 = -159000.00382
    else:
        vel_v1 = -154800.004

    # 0.01% agreement is good enough across different libraries.
    # RA can be off by 360 degrees and still be valid.
    if modname == 'astlib':
        with pytest.raises(NotImplementedError):
            img.wcs.datapt_to_wcspt(xy_v1, naxispath=nxp)
    else:
        radec = img.wcs.datapt_to_wcspt(xy_v1, naxispath=nxp)
        assert (np.allclose(radec[:, 0], radec_deg_v1[:, 0]) or
                np.allclose(radec[:, 0], radec_deg_v1[:, 0] + 360))
        assert_allclose(radec[:, 1], radec_deg_v1[:, 1], rtol=1e-4)
        if modname == 'starlink':
            with pytest.raises(IndexError):
                radec[:, 2]
        else:
            assert_allclose(radec[:, 2], vel_v1, rtol=1e-4)

    if modname == 'astlib':
        with pytest.raises(NotImplementedError):
            img.wcs.wcspt_to_datapt(radec_deg_v1, naxispath=nxp)
    else:
        xy = img.wcs.wcspt_to_datapt(radec_deg_v1, naxispath=nxp)
        assert_allclose(xy[:, :2], xy_v1, atol=3e-6)
        if modname == 'kapteyn':
            assert_allclose(xy[:, 2], 36.85715, atol=3e-6)

    if modname == 'astlib':
        with pytest.raises(NotImplementedError):
            img.wcs.datapt_to_system([(0, 0, 0), (120, 100, 0)],
                                     system='galactic')
    elif modname == 'kapteyn':
        with pytest.raises(Exception):
            img.wcs.datapt_to_system([(0, 0, 0), (120, 100, 0)],
                                     system='galactic')
    else:
        gal = img.wcs.datapt_to_system([(0, 0, 0), (120, 100, 0)],
                                       system='galactic')
        if modname in ('astropy', 'astropy_ape14'):
            assert_allclose(gal.l.degree, gal_v1[:, 0])
            assert_allclose(gal.b.degree, gal_v1[:, 1])
        else:
            assert_allclose(gal, gal_v1, rtol=1e-4)


def test_fixheader():
    w = wcsmod.common.BaseWCS(_logger)
    w.header = {'SIMPLE': True, 'CUNIT1': 'degree', 'CUNIT2': 'Degree'}
    w.fix_bad_headers()
    assert w.get_keyword('SIMPLE')
    assert w.get_keywords('CUNIT1', 'CUNIT2') == ['deg', 'deg']

    with pytest.raises(wcsmod.common.WCSError):
        w.datapt_to_system((0, 0))


@pytest.mark.parametrize('val', ['degr', 'blah'])
def test_choose_coord_units(val):
    assert wcsmod.common.choose_coord_units({'CUNIT1': val}) == 'degree'


@pytest.mark.parametrize(
    ('hdr', 'val'),
    [({'RA': 0, 'EQUINOX': 1983.9}, 'fk4'),
     ({'RA': 0, 'EQUINOX': 1984.0}, 'fk5'),
     ({'RA': 0}, 'icrs'),
     ({}, 'raw'),
     ({'CTYPE1': 'GLON-TAN'}, 'galactic'),
     ({'CTYPE1': 'ELON-TAN'}, 'ecliptic'),
     ({'CTYPE1': 'RA---TAN', 'EQUINOX': 1983.9}, 'fk4'),
     ({'CTYPE1': 'RA---TAN', 'EQUINOX': 1984.0}, 'fk5'),
     ({'CTYPE1': 'RA---TAN'}, 'icrs'),
     ({'CTYPE1': 'RA---TAN', 'RADECSYS': 'foo'}, 'foo'),
     ({'CTYPE1': 'RA---TAN', 'RADESYS': 'bar'}, 'bar'),
     ({'CTYPE1': 'HPLN-TAN'}, 'helioprojective'),
     ({'CTYPE1': 'HGLT-TAN'}, 'heliographicstonyhurst'),
     ({'CTYPE1': 'PIXEL'}, 'pixel'),
     ({'CTYPE1': 'LINEAR'}, 'pixel'),
     ({'CTYPE1': 'foo'}, 'icrs')])
def test_get_coord_sys_name(hdr, val):
    assert wcsmod.common.get_coord_system_name(hdr) == val

# END



#
# AstroTable.py -- Abstraction of an generic astro table.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np
from astropy.table import Table

from ginga.BaseImage import ViewerObjectBase, Header
from ginga.util import wcsmod


class TableError(Exception):
    pass


class AstroTableHeader(Header):
    pass


class AstroTable(ViewerObjectBase):
    """Abstraction of an astronomical data (table).

    .. note:: This module is NOT thread-safe!

    """
    # class variables for WCS can be set
    wcsClass = None
    ioClass = None

    @classmethod
    def set_wcsClass(cls, klass):
        cls.wcsClass = klass

    @classmethod
    def set_ioClass(cls, klass):
        cls.ioClass = klass

    def __init__(self, data_ap=None, metadata=None, logger=None, name=None,
                 wcsclass=wcsClass, ioclass=ioClass):

        ViewerObjectBase.__init__(self, logger=logger, metadata=metadata,
                                  name=name)

        self._data = data_ap
        self.naxispath = []

        # wcsclass specifies a pluggable WCS module
        if wcsclass is None:
            wcsclass = wcsmod.WCS
        self.wcs = wcsclass(self.logger)

        # ioclass specifies a pluggable IO module
        if ioclass is None:
            from ginga.util import io_fits
            ioclass = io_fits.fitsLoaderClass
        self.io = ioclass(self.logger)

        # TODO: How to handle table with WCS data? For example, spectrum
        #       table may store dispersion solution as WCS.
        #if metadata is not None:
        #    header = self.get_header()
        #    self.wcs.load_header(header)

    @property
    def rows(self):
        tab_a = self._get_data()
        return len(tab_a)

    @property
    def columns(self):
        tab_a = self._get_data()
        return len(tab_a.colnames)

    def get_size(self):
        return (self.columns, self.rows)

    def get_data(self):
        return self._data

    def _get_data(self):
        return self._data

    def get_header(self, create=True, **kwargs):
        try:
            # By convention, the fits header is stored in a dictionary
            # under the metadata keyword 'header'
            displayhdr = self.metadata['header']

        except KeyError as e:
            if not create:
                raise e
            displayhdr = AstroTableHeader()
            self.metadata['header'] = displayhdr

        return displayhdr

    def set_data(self, data_ap, metadata=None):
        """Use this method to SHARE (not copy) the incoming table.
        """
        self._data = data_ap

        if metadata:
            self.update_metadata(metadata)

        self.make_callback('modified')

    def clear_all(self):
        # clear metadata
        super(AstroTable, self).clear_all()

        # unreference data
        self._data = None

    def get_minmax(self, noinf=False):
        # TODO: what should this mean for a table?
        return (0, 0)

    def load_hdu(self, hdu, fobj=None, **kwargs):
        self.clear_metadata()

        # These keywords might be provided but not used.
        if 'inherit_primary_header' in kwargs:
            kwargs.pop('inherit_primary_header')

        ahdr = self.get_header()

        if 'format' not in kwargs:
            kwargs['format'] = 'fits'

        try:
            self._data = Table.read(hdu, **kwargs)
            ahdr.update(hdu.header)

        except Exception as e:
            self.logger.error("Error reading table from hdu: {0}".format(
                str(e)))

        # TODO: Try to make a wcs object on the header
        #self.wcs.load_header(hdu.header, fobj=fobj)

    def load_file(self, filespec, **kwargs):
        if self.io is None:
            raise TableError("No IO loader defined")

        self.io.load_file(filespec, dstobj=self, **kwargs)

    def get_thumbnail(self, length):
        thumb_np = np.eye(length)
        return thumb_np

#END



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
# viewer.py -- maintains information about registered viewers for data
#

from ginga.misc import Bunch

# this holds registered viewers
viewer_db = {}


def register_viewer(vclass):
    """Register a channel viewer.

    Parameters
    ----------
    vclass : :py:class
        The class of the viewer.
    """
    global viewer_db
    viewer_db[vclass.vname] = Bunch.Bunch(name=vclass.vname,
                                          vclass=vclass,
                                          priority=0)


def get_viewers(dataobj):
    """Returns a list of viewers that are registered that can view `dataobj`,
    sorted by priority.
    """
    res = []
    for vinfo in viewer_db.values():
        vclass = vinfo.vclass
        if not hasattr(vclass, 'viewable'):
            continue
        try:
            tf = vclass.viewable(dataobj)
            if tf:
                res.append(vinfo)
        except Exception as e:
            pass

    res.sort(key=lambda v: v.priority)
    return res


def get_priority_viewers(dataobj):
    """Returns a list of viewers that are registered that can view `dataobj`,
    with all but the best (matching) priority removed.
    """
    res = get_viewers(dataobj)
    if len(res) == 0:
        return res

    # if there is more than one possible viewer, return the list of
    # those that have the best (i.e. lowest) equal priority
    priority = res[0].priority
    return [vinfo for vinfo in res if vinfo.priority <= priority]


def get_viewer_names(dataobj):
    """Returns a list of viewer names that are registered that
    can view `dataobj`.
    """
    return [vinfo.name for vinfo in get_viewers(dataobj)]



#
# Widgets.py -- wrapped Qt widgets and convenience functions
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import os.path
from functools import reduce

from ginga.qtw.QtHelp import (QtGui, QtCore, QTextCursor, QIcon, QPixmap,
                              QImage, QCursor, QFont, have_pyqt4)
from ginga.qtw import QtHelp

from ginga.misc import Callback, Bunch, LineHistory
import ginga.icons

has_webkit = False
try:
    from ginga.qtw.QtHelp import QWebView  # noqa
    has_webkit = True
except ImportError:
    pass

__all__ = ['WidgetError', 'WidgetBase', 'TextEntry', 'TextEntrySet',
           'GrowingTextEdit', 'TextArea', 'Label', 'Button', 'ComboBox',
           'SpinBox', 'Slider', 'ScrollBar', 'CheckBox', 'ToggleButton',
           'RadioButton', 'Image', 'ProgressBar', 'StatusBar', 'TreeView',
           'WebView', 'ContainerBase', 'Box', 'HBox', 'VBox', 'Frame',
           'Expander', 'TabWidget', 'StackWidget', 'MDIWidget', 'ScrollArea',
           'Splitter', 'GridBox', 'ToolbarAction', 'Toolbar', 'MenuAction',
           'Menu', 'Menubar', 'TopLevelMixin', 'TopLevel', 'Application',
           'Dialog', 'SaveDialog', 'DragPackage',
           'name_mangle', 'make_widget', 'hadjust', 'build_info', 'wrap',
           'has_webkit']

# path to our icons
icondir = os.path.split(ginga.icons.__file__)[0]


class WidgetError(Exception):
    """For errors thrown in this module."""
    pass


_app = None


# BASE

class WidgetBase(Callback.Callbacks):

    def __init__(self):
        super(WidgetBase, self).__init__()

        self.widget = None
        self.changed = False
        # external data can be attached here
        self.extdata = Bunch.Bunch()

    def get_widget(self):
        return self.widget

    def set_tooltip(self, text):
        self.widget.setToolTip(text)

    def get_enabled(self):
        self.widget.isEnabled()

    def set_enabled(self, tf):
        self.widget.setEnabled(tf)

    def get_size(self):
        wd, ht = self.widget.width(), self.widget.height()
        return (wd, ht)

    def get_pos(self):
        x, y = self.widget.x(), self.widget.y()
        return (x, y)

    def get_app(self):
        return _app

    def delete(self):
        self.widget.deleteLater()

    def focus(self):
        self.widget.activateWindow()
        self.widget.setFocus()
        # self.widget.raise_()

    def resize(self, width, height):
        self.widget.resize(width, height)

    def show(self):
        self.widget.show()

    def hide(self):
        self.widget.hide()

    def get_font(self, font_family, point_size):
        font = QtHelp.get_font(font_family, point_size)
        return font

    def cfg_expand(self, horizontal=0, vertical=0):
        h_policy = QtGui.QSizePolicy.Policy(horizontal)
        v_policy = QtGui.QSizePolicy.Policy(vertical)
        self.widget.setSizePolicy(QtGui.QSizePolicy(h_policy, v_policy))


# BASIC WIDGETS

class TextEntry(WidgetBase):
    def __init__(self, text='', editable=True):
        super(TextEntry, self).__init__()

        self.widget = QtGui.QLineEdit()
        self.widget.setText(text)
        self.widget.setReadOnly(not editable)
        self.widget.returnPressed.connect(self._cb_redirect)
        self.widget.keyPressEvent_org = self.widget.keyPressEvent
        self.widget.keyPressEvent = self._key_press_event

        self.history = LineHistory.LineHistory()

        self.enable_callback('activated')

    def _cb_redirect(self, *args):
        self.history.append(self.get_text())
        self.make_callback('activated')

    def _key_press_event(self, event):
        keycode = event.key()
        if keycode in [QtCore.Qt.Key_Up]:
            try:
                self.set_text(self.history.prev())
            except ValueError:
                pass
        elif keycode in [QtCore.Qt.Key_Down]:
            try:
                self.set_text(self.history.next())
            except ValueError:
                pass
        else:
            self.widget.keyPressEvent_org(event)

    def get_text(self):
        return self.widget.text()

    def set_text(self, text):
        self.widget.setText(text)

    def set_editable(self, tf):
        self.widget.setReadOnly(not tf)

    def set_font(self, font, size=10):
        if not isinstance(font, QFont):
            font = self.get_font(font, size)
        self.widget.setFont(font)

    def set_length(self, numchars):
        # this is only supposed to set the visible length (but Qt doesn't
        # really have a good way to do that)
        # self.widget.setMaxLength(numchars)
        pass


class TextEntrySet(WidgetBase):
    def __init__(self, text='', editable=True):
        super(TextEntrySet, self).__init__()

        self.widget = QtHelp.HBox()
        self.entry = QtGui.QLineEdit()
        self.entry.setText(text)
        self.entry.setReadOnly(not editable)
        layout = self.widget.layout()
        layout.addWidget(self.entry, stretch=1)
        self.btn = QtGui.QPushButton('Set')
        self.entry.returnPressed.connect(self._cb_redirect)
        self.btn.clicked.connect(self._cb_redirect)
        layout.addWidget(self.btn, stretch=0)

        self.enable_callback('activated')

    def _cb_redirect(self, *args):
        self.make_callback('activated')

    def get_text(self):
        return self.entry.text()

    def set_text(self, text):
        self.entry.setText(text)

    def set_editable(self, tf):
        self.entry.setReadOnly(not tf)

    def set_font(self, font, size=10):
        if not isinstance(font, QFont):
            font = self.get_font(font, size)
        self.widget.setFont(font)

    def set_length(self, numchars):
        # this is only supposed to set the visible length (but Qt doesn't
        # really have a good way to do that)
        # self.widget.setMaxLength(numchars)
        pass

    def set_enabled(self, tf):
        super(TextEntrySet, self).set_enabled(tf)
        self.entry.setEnabled(tf)


class GrowingTextEdit(QtGui.QTextEdit):

    def __init__(self, *args, **kwargs):
        super(GrowingTextEdit, self).__init__(*args, **kwargs)
        self.document().documentLayout().documentSizeChanged.connect(
            self.sizeChange)
        self.heightMin = 0
        self.heightMax = 65000

    def sizeChange(self):
        docHeight = self.document().size().height()
        # add some margin to prevent auto scrollbars
        docHeight += 20
        if self.heightMin <= docHeight <= self.heightMax:
            self.setMaximumHeight(docHeight)


class TextArea(WidgetBase):
    def __init__(self, wrap=False, editable=False):
        super(TextArea, self).__init__()

        # tw = QtGui.QTextEdit()
        tw = GrowingTextEdit()
        tw.setReadOnly(not editable)
        if wrap:
            tw.setLineWrapMode(QtGui.QTextEdit.WidgetWidth)
        else:
            tw.setLineWrapMode(QtGui.QTextEdit.NoWrap)
        self.widget = tw

    def append_text(self, text, autoscroll=True):
        if text.endswith('\n'):
            text = text[:-1]
        self.widget.append(text)
        if not autoscroll:
            return

        self.widget.moveCursor(QTextCursor.End)
        self.widget.moveCursor(QTextCursor.StartOfLine)
        self.widget.ensureCursorVisible()

    def get_text(self):
        return self.widget.document().toPlainText()

    def clear(self):
        self.widget.clear()

    def set_text(self, text):
        self.clear()
        self.append_text(text)

    def set_editable(self, tf):
        self.widget.setReadOnly(not tf)

    def set_limit(self, numlines):
        # self.widget.setMaximumBlockCount(numlines)
        pass

    def set_font(self, font, size=10):
        if not isinstance(font, QFont):
            font = self.get_font(font, size)
        self.widget.setCurrentFont(font)

    def set_wrap(self, tf):
        if tf:
            self.widget.setLineWrapMode(QtGui.QTextEdit.WidgetWidth)
        else:
            self.widget.setLineWrapMode(QtGui.QTextEdit.NoWrap)


class Label(WidgetBase):
    def __init__(self, text='', halign='left', style='normal', menu=None):
        super(Label, self).__init__()

        lbl = QtGui.QLabel(text)
        if halign == 'left':
            lbl.setAlignment(QtCore.Qt.AlignLeft)
        elif halign == 'center':
            lbl.setAlignment(QtCore.Qt.AlignHCenter)
        elif halign == 'right':
            lbl.setAlignment(QtCore.Qt.AlignRight)

        self.widget = lbl
        lbl.mousePressEvent = self._cb_redirect
        lbl.mouseReleaseEvent = self._cb_redirect2

        if style == 'clickable':
            lbl.setSizePolicy(QtGui.QSizePolicy.Minimum,
                              QtGui.QSizePolicy.Minimum)
            lbl.setFrameStyle(QtGui.QFrame.Box | QtGui.QFrame.Raised)

        if menu is not None:
            lbl.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            menu_w = menu.get_widget()

            def on_context_menu(point):
                menu_w.exec_(lbl.mapToGlobal(point))

            lbl.customContextMenuRequested.connect(on_context_menu)

        # Enable highlighting for copying
        # lbl.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)

        self.enable_callback('activated')
        self.enable_callback('released')

    def _cb_redirect(self, event):
        buttons = event.buttons()
        if buttons & QtCore.Qt.LeftButton:
            self.make_callback('activated')

    def _cb_redirect2(self, event):
        buttons = event.buttons()
        if buttons & QtCore.Qt.LeftButton:
            self.make_callback('released')

    def get_text(self):
        return self.widget.text()

    def set_text(self, text):
        self.widget.setText(text)

    def set_font(self, font, size=10):
        if not isinstance(font, QFont):
            font = self.get_font(font, size)
        self.widget.setFont(font)

    def set_color(self, fg=None, bg=None):
        self.widget.setStyleSheet(
            "QLabel { background-color: %s; color: %s; }" % (bg, fg))


class Button(WidgetBase):
    def __init__(self, text=''):
        super(Button, self).__init__()

        self.widget = QtGui.QPushButton(text)
        self.widget.clicked.connect(self._cb_redirect)

        self.enable_callback('activated')

    def _cb_redirect(self, *args):
        self.make_callback('activated')


class ComboBox(WidgetBase):
    def __init__(self, editable=False):
        super(ComboBox, self).__init__()

        self.widget = QtHelp.ComboBox()
        self.widget.setSizeAdjustPolicy(QtGui.QComboBox.AdjustToContents)
        self.widget.setEditable(editable)
        self.widget.activated.connect(self._cb_redirect)

        self.enable_callback('activated')

    def _cb_redirect(self):
        idx = self.widget.currentIndex()
        self.make_callback('activated', idx)

    def insert_alpha(self, text):
        index = 0
        while True:
            itemText = self.widget.itemText(index)
            if len(itemText) == 0:
                break
            if itemText > text:
                self.widget.insertItem(index, text)
                return
            index += 1
        self.widget.addItem(text)

    def delete_alpha(self, text):
        index = self.widget.findText(text)
        self.widget.removeItem(index)

    def get_alpha(self, idx):
        return self.widget.itemText(idx)

    def clear(self):
        self.widget.clear()

    def set_text(self, text):
        index = self.widget.findText(text)
        if index >= 0:
            self.set_index(index)
        else:
            self.widget.setEditText(text)

    # to be deprecated someday
    show_text = set_text

    def get_text(self):
        idx = self.get_index()
        return self.get_alpha(idx)

    def append_text(self, text):
        self.widget.addItem(text)

    def set_index(self, index):
        self.widget.setCurrentIndex(index)

    def get_index(self):
        return self.widget.currentIndex()


class SpinBox(WidgetBase):
    def __init__(self, dtype=int):
        super(SpinBox, self).__init__()

        if dtype == float:
            w = QtGui.QDoubleSpinBox()
        else:
            w = QtGui.QSpinBox()
        w.valueChanged.connect(self._cb_redirect)
        # should values wrap around
        w.setWrapping(False)
        self.widget = w

        self.enable_callback('value-changed')

    def _cb_redirect(self, val):
        if self.changed:
            self.changed = False
            return
        self.make_callback('value-changed', val)

    def get_value(self):
        return self.widget.value()

    def set_value(self, val):
        self.changed = True
        self.widget.setValue(val)

    def set_decimals(self, num):
        if hasattr(self.widget, 'setDecimals'):
            # only for QDoubleSpinBox
            self.widget.setDecimals(num)

    def set_limits(self, minval, maxval, incr_value=1):
        adj = self.widget
        adj.setRange(minval, maxval)
        adj.setSingleStep(incr_value)


class Slider(WidgetBase):
    def __init__(self, orientation='horizontal', dtype=int, track=False):
        super(Slider, self).__init__()

        # NOTE: parameter dtype is ignored for now for Qt

        if orientation == 'horizontal':
            w = QtGui.QSlider(QtCore.Qt.Horizontal)
            w.setTickPosition(QtGui.QSlider.TicksBelow)
        else:
            w = QtGui.QSlider(QtCore.Qt.Vertical)
            w.setTickPosition(QtGui.QSlider.TicksRight)
        # w.setTickPosition(QtGui.QSlider.NoTicks)
        # this controls whether the callbacks are made *as the user
        # moves the slider* or afterwards
        w.setTracking(track)
        self.widget = w
        w.valueChanged.connect(self._cb_redirect)

        self.enable_callback('value-changed')

    def _cb_redirect(self, val):
        # It appears that Qt uses set_value() to set the value of the
        # slider when it is dragged, so we cannot use the usual method
        # of setting a hidden "changed" variable to suppress the callback
        # when setting the value programmatically.
        # if self.changed:
        #     self.changed = False
        #     return
        self.make_callback('value-changed', val)

    def get_value(self):
        return self.widget.value()

    def set_value(self, val):
        self.changed = True
        self.widget.setValue(val)

    def set_tracking(self, tf):
        self.widget.setTracking(tf)

    def set_limits(self, minval, maxval, incr_value=1):
        adj = self.widget
        adj.setRange(minval, maxval)
        adj.setSingleStep(incr_value)


class ScrollBar(WidgetBase):
    def __init__(self, orientation='horizontal'):
        super(ScrollBar, self).__init__()

        if orientation == 'horizontal':
            self.widget = QtGui.QScrollBar(QtCore.Qt.Horizontal)
        else:
            self.widget = QtGui.QScrollBar(QtCore.Qt.Vertical)
        self.widget.valueChanged.connect(self._cb_redirect)

        self.enable_callback('activated')

    def _cb_redirect(self):
        val = self.widget.value()
        self.make_callback('activated', val)


class CheckBox(WidgetBase):
    def __init__(self, text=''):
        super(CheckBox, self).__init__()

        self.widget = QtGui.QCheckBox(text)
        self.widget.stateChanged.connect(self._cb_redirect)

        self.enable_callback('activated')

    def _cb_redirect(self, *args):
        val = self.get_state()
        self.make_callback('activated', val)

    def set_state(self, tf):
        self.widget.setChecked(tf)

    def get_state(self):
        val = self.widget.checkState()
        # returns 0 (unchecked) or 2 (checked)
        return (val != 0)


class ToggleButton(WidgetBase):
    def __init__(self, text=''):
        super(ToggleButton, self).__init__()

        self.widget = QtGui.QPushButton(text)
        self.widget.setCheckable(True)
        self.widget.clicked.connect(self._cb_redirect)

        self.enable_callback('activated')

    def _cb_redirect(self, val):
        self.make_callback('activated', val)

    def set_state(self, tf):
        self.widget.setChecked(tf)

    def get_state(self):
        return self.widget.isChecked()


class RadioButton(WidgetBase):
    def __init__(self, text='', group=None):
        super(RadioButton, self).__init__()

        self.widget = QtGui.QRadioButton(text)
        self.widget.toggled.connect(self._cb_redirect)

        self.enable_callback('activated')

    def _cb_redirect(self, val):
        if self.changed:
            self.changed = False
            return
        self.make_callback('activated', val)

    def set_state(self, tf):
        if self.widget.isChecked() != tf:
            # toggled only fires when the value is toggled
            self.changed = True
            self.widget.setChecked(tf)

    def get_state(self):
        return self.widget.isChecked()


class Image(WidgetBase):
    def __init__(self, native_image=None, style='normal', menu=None):
        super(Image, self).__init__()

        lbl = QtGui.QLabel()
        self.widget = lbl
        if native_image is not None:
            self._set_image(native_image)

        lbl.mousePressEvent = self._cb_redirect

        if style == 'clickable':
            lbl.setSizePolicy(QtGui.QSizePolicy.Minimum,
                              QtGui.QSizePolicy.Minimum)
            # lbl.setFrameStyle(QtGui.QFrame.Box | QtGui.QFrame.Raised)

        if menu is not None:
            lbl.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            menu_w = menu.get_widget()

            def on_context_menu(point):
                menu_w.exec_(lbl.mapToGlobal(point))

            lbl.customContextMenuRequested.connect(on_context_menu)

        self.enable_callback('activated')

    def _cb_redirect(self, event):
        buttons = event.buttons()
        if buttons & QtCore.Qt.LeftButton:
            self.make_callback('activated')

    def _set_image(self, native_image):
        pixmap = QPixmap.fromImage(native_image)
        self.widget.setPixmap(pixmap)

    def load_file(self, img_path, format=None):
        pixmap = QPixmap()
        pixmap.load(img_path, format=format)
        self.widget.setPixmap(pixmap)


class ProgressBar(WidgetBase):
    def __init__(self):
        super(ProgressBar, self).__init__()

        w = QtGui.QProgressBar()
        w.setRange(0, 100)
        w.setTextVisible(True)
        self.widget = w

    def set_value(self, pct):
        self.widget.setValue(int(pct * 100.0))


class StatusBar(WidgetBase):
    def __init__(self):
        super(StatusBar, self).__init__()

        sbar = QtGui.QStatusBar()
        self.widget = sbar

    def set_message(self, msg_str, duration=10.0):
        # remove message in about `duration` seconds
        self.widget.showMessage(msg_str, int(duration * 1000))


class TreeWidgetItem(QtGui.QTreeWidgetItem):
    """A hack to subclass QTreeWidgetItem to enable sorting by numbers
    in a field.
    """
    def __init__(self, *args, **kwargs):
        QtGui.QTreeWidgetItem.__init__(self, *args, **kwargs)

    def __lt__(self, otherItem):
        column = self.treeWidget().sortColumn()
        try:
            return float(self.text(column)) < float(otherItem.text(column))

        except ValueError:
            return self.text(column) < otherItem.text(column)


class TreeView(WidgetBase):
    def __init__(self, auto_expand=False, sortable=False,
                 selection='single', use_alt_row_color=False,
                 dragable=False):
        super(TreeView, self).__init__()

        self.auto_expand = auto_expand
        self.sortable = sortable
        self.dragable = dragable
        self.selection = selection
        self.levels = 1
        self.leaf_key = None
        self.leaf_idx = 0
        self.columns = []
        self.datakeys = []
        # shadow index
        self.shadow = {}

        tv = QtGui.QTreeWidget()
        self.widget = tv
        tv.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
        if selection == 'multiple':
            tv.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)
        tv.setAlternatingRowColors(use_alt_row_color)
        tv.itemDoubleClicked.connect(self._cb_redirect)
        tv.itemSelectionChanged.connect(self._selection_cb)
        if self.dragable:
            tv.setDragEnabled(True)
            tv.startDrag = self._start_drag

        for cbname in ('selected', 'activated', 'drag-start'):
            self.enable_callback(cbname)

    def setup_table(self, columns, levels, leaf_key):
        self.clear()

        self.columns = columns
        self.levels = levels
        self.leaf_key = leaf_key
        treeview = self.widget
        treeview.setColumnCount(len(columns))
        treeview.setSortingEnabled(self.sortable)

        # speeds things up a bit
        treeview.setUniformRowHeights(True)

        # create the column headers
        if not isinstance(columns[0], str):
            # columns specifies a mapping
            headers = [col[0] for col in columns]
            datakeys = [col[1] for col in columns]
        else:
            headers = datakeys = columns

        self.datakeys = datakeys
        self.leaf_idx = datakeys.index(self.leaf_key)

        if self.sortable:
            # Sort increasing by default
            treeview.sortByColumn(self.leaf_idx, QtCore.Qt.AscendingOrder)

        treeview.setHeaderLabels(headers)

    def set_tree(self, tree_dict):
        self.clear()
        self.add_tree(tree_dict)

    def add_tree(self, tree_dict):

        if self.sortable:
            self.widget.setSortingEnabled(False)

        for key in tree_dict:
            self._add_subtree(1, self.shadow,
                              self.widget, key, tree_dict[key])

        if self.sortable:
            self.widget.setSortingEnabled(True)

        # User wants auto expand?
        if self.auto_expand:
            self.widget.expandAll()

    def _add_subtree(self, level, shadow, parent_item, key, node):

        if level >= self.levels:
            # leaf node
            values = ['' if _key == 'icon' else str(node[_key])
                      for _key in self.datakeys]
            try:
                bnch = shadow[key]
                item = bnch.item
                # TODO: update leaf item

            except KeyError:
                # new item
                item = TreeWidgetItem(parent_item, values)
                if level == 1:
                    parent_item.addTopLevelItem(item)
                else:
                    parent_item.addChild(item)

                shadow[key] = Bunch.Bunch(node=node, item=item, terminal=True)

                # hack for adding an image to a table
                # TODO: add types for columns
                if 'icon' in node:
                    i = self.datakeys.index('icon')
                    item.setIcon(i, node['icon'])

                # mark cell as non-editable
                item.setFlags(item.flags() & ~QtCore.Qt.ItemIsEditable)

        else:
            try:
                # node already exists
                bnch = shadow[key]
                item = bnch.item
                d = bnch.node

            except KeyError:
                # new node
                item = TreeWidgetItem(parent_item, [str(key)])
                if level == 1:
                    parent_item.addTopLevelItem(item)
                else:
                    parent_item.addChild(item)
                d = {}
                shadow[key] = Bunch.Bunch(node=d, item=item, terminal=False)

            # recurse for non-leaf interior node
            for key in node:
                self._add_subtree(level + 1, d, item, key, node[key])

    def _selection_cb(self):
        res_dict = self.get_selected()
        self.make_callback('selected', res_dict)

    def _cb_redirect(self, item):
        res_dict = {}
        self._get_item(res_dict, item)
        self.make_callback('activated', res_dict)

    def _get_path(self, item):
        if item is None:
            return []

        if item.childCount() == 0:
            path_rest = self._get_path(item.parent())
            myname = item.text(self.leaf_idx)
            path_rest.append(myname)
            return path_rest

        myname = item.text(0)
        path_rest = self._get_path(item.parent())
        path_rest.append(myname)
        return path_rest

    def _get_item(self, res_dict, item):
        # from the QTreeViewItem `item`, return the item via a path
        # in the dictionary `res_dict`
        path = self._get_path(item)
        d, s = res_dict, self.shadow
        for name in path[:-1]:
            d = d.setdefault(name, {})
            s = s[name].node

        dst_key = path[-1]
        try:
            d[dst_key] = s[dst_key].node
        except KeyError:
            d[dst_key] = None

    def get_selected(self):
        items = list(self.widget.selectedItems())
        res_dict = {}
        for item in items:
            if item.childCount() > 0:
                # only leaf nodes can be selected
                continue
            self._get_item(res_dict, item)
        return res_dict

    def clear(self):
        self.widget.clear()
        self.shadow = {}

    def clear_selection(self):
        self.widget.clearSelection()

    def _path_to_item(self, path):
        s = self.shadow
        for name in path[:-1]:
            s = s[name].node
        item = s[path[-1]].item
        return item

    def select_path(self, path, state=True):
        item = self._path_to_item(path)
        item.setSelected(state)

    def highlight_path(self, path, onoff, font_color='green'):
        item = self._path_to_item(path)

        # A little painfully inefficient, can we do better than this?
        font = QtHelp.QFont()
        if not onoff:
            color = QtHelp.QColor('black')
        else:
            font.setBold(True)
            color = QtHelp.QColor(font_color)
        brush = QtHelp.QBrush(color)

        for i in range(item.columnCount()):
            item.setForeground(i, brush)
            item.setFont(i, font)

    def scroll_to_path(self, path):
        # TODO: this doesn't give an error, but does not seem to be
        # working as the API indicates
        item = self._path_to_item(path)
        row = self.widget.indexOfTopLevelItem(item)
        midx = self.widget.indexAt(QtCore.QPoint(row, 0))
        self.widget.scrollTo(midx, QtGui.QAbstractItemView.PositionAtCenter)

    def sort_on_column(self, i):
        self.widget.sortByColumn(i, QtCore.Qt.AscendingOrder)

    def set_column_width(self, i, width):
        self.widget.setColumnWidth(i, width)

    def set_column_widths(self, lwidths):
        for i, width in enumerate(lwidths):
            if width is not None:
                self.set_column_width(i, width)

    def set_optimal_column_widths(self):
        for i in range(len(self.columns)):
            self.widget.resizeColumnToContents(i)

    def _start_drag(self, event):
        res_dict = self.get_selected()
        drag_pkg = DragPackage(self.widget)
        self.make_callback('drag-start', drag_pkg, res_dict)
        drag_pkg.start_drag()


class WebView(WidgetBase):
    def __init__(self):
        if not has_webkit:
            raise NotImplementedError("Missing webkit")

        super(WebView, self).__init__()
        self.widget = QWebView()

    def load_url(self, url):
        self.widget.load(QtCore.QUrl(url))

    def load_html_string(self, html_string):
        self.widget.setHtml(html_string)

    def go_back(self):
        self.widget.back()

    def go_forward(self):
        self.widget.forward()

    def reload_page(self):
        self.widget.reload()

    def stop_loading(self):
        self.widget.stop()


# CONTAINERS

class ContainerBase(WidgetBase):
    def __init__(self):
        super(ContainerBase, self).__init__()
        self.children = []

        for name in ['widget-added', 'widget-removed']:
            self.enable_callback(name)

    def add_ref(self, ref):
        # TODO: should this be a weakref?
        self.children.append(ref)

    def _remove(self, childw, delete=False):
        layout = self.widget.layout()
        if layout is not None:
            layout.removeWidget(childw)

        childw.setParent(None)
        if delete:
            childw.deleteLater()

    def remove(self, child, delete=False):
        if child not in self.children:
            raise ValueError("Widget is not a child of this container")
        self.children.remove(child)

        self._remove(child.get_widget(), delete=delete)
        self.make_callback('widget-removed', child)

    def remove_all(self, delete=False):
        for w in list(self.children):
            self.remove(w, delete=delete)

    def get_children(self):
        return self.children

    def num_children(self):
        return len(self.children)

    def _get_native_children(self):
        return [child.get_widget() for child in self.children]

    def _get_native_index(self, nchild):
        l = self._get_native_children()
        try:
            return l.index(nchild)
        except (IndexError, ValueError) as e:
            return -1

    def _native_to_child(self, nchild):
        idx = self._get_native_index(nchild)
        if idx < 0:
            return None
        return self.children[idx]

    def set_margins(self, left, right, top, bottom):
        layout = self.widget.layout()
        layout.setContentsMargins(left, right, top, bottom)

    def set_border_width(self, pix):
        layout = self.widget.layout()
        layout.setContentsMargins(pix, pix, pix, pix)


class Box(ContainerBase):
    def __init__(self, orientation='horizontal'):
        super(Box, self).__init__()

        self.widget = QtGui.QWidget()
        self.orientation = orientation
        if orientation == 'horizontal':
            self.layout = QtGui.QHBoxLayout()
        else:
            self.layout = QtGui.QVBoxLayout()

        # because of ridiculous defaults
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.widget.setLayout(self.layout)

    def add_widget(self, child, stretch=0.0):
        self.add_ref(child)
        child_w = child.get_widget()
        self.layout.addWidget(child_w, stretch=stretch)
        self.make_callback('widget-added', child)

    def set_spacing(self, val):
        self.layout.setSpacing(val)


class HBox(Box):
    def __init__(self):
        super(HBox, self).__init__(orientation='horizontal')


class VBox(Box):
    def __init__(self):
        super(VBox, self).__init__(orientation='vertical')


class Frame(ContainerBase):
    def __init__(self, title=None):
        super(Frame, self).__init__()

        self.widget = QtGui.QFrame()
        self.widget.setFrameStyle(QtGui.QFrame.Box | QtGui.QFrame.Raised)
        vbox = QtGui.QVBoxLayout()
        self.layout = vbox
        # because of ridiculous defaults
        vbox.setContentsMargins(2, 2, 2, 2)
        self.widget.setLayout(vbox)
        if title:
            lbl = QtGui.QLabel(title)
            lbl.setAlignment(QtCore.Qt.AlignHCenter)
            # lbl.setAlignment(QtCore.Qt.AlignLeft)
            vbox.addWidget(lbl, stretch=0)
            self.label = lbl
        else:
            self.label = None

    def set_widget(self, child, stretch=1):
        self.remove_all()
        self.add_ref(child)
        self.widget.layout().addWidget(child.get_widget(), stretch=stretch)


# Qt custom expander widget
# See http://stackoverflow.com/questions/10364589/equivalent-of-gtks-expander-in-pyqt4  # noqa
#
class Expander(ContainerBase):
    r_arrow = None
    d_arrow = None

    # Note: add 'text-align: left;' if you want left adjusted labels
    widget_style = """
    QPushButton { margin: 1px,1px,1px,1px; padding: 0px;
                  border-width: 0px; border-style: solid; }
    """

    def __init__(self, title=''):
        super(Expander, self).__init__()

        # Qt doesn't seem to like it (segfault) if we actually construct
        # these icons in the class variable declarations
        if Expander.r_arrow is None:
            Expander.r_arrow = QtHelp.get_icon(
                os.path.join(icondir, 'triangle-right-48.png'), size=(12, 12))
        if Expander.d_arrow is None:
            Expander.d_arrow = QtHelp.get_icon(
                os.path.join(icondir, 'triangle-down-48.png'), size=(12, 12))

        self.widget = QtGui.QWidget()
        vbox = QtGui.QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.setSpacing(0)
        self.layout = vbox

        self.toggle = QtGui.QPushButton(Expander.r_arrow, title)
        self.toggle.setStyleSheet(Expander.widget_style)
        # self.toggle.setCheckable(True)
        self.toggle.clicked.connect(self._toggle_widget)

        vbox.addWidget(self.toggle, stretch=0)
        self.widget.setLayout(vbox)

    def set_widget(self, child, stretch=1):
        self.remove_all()
        self.add_ref(child)
        child_w = child.get_widget()
        self.widget.layout().addWidget(child_w, stretch=stretch)
        child_w.setVisible(False)

    def _toggle_widget(self):
        child = self.get_children()[0]
        child_w = child.get_widget()
        # if self.toggle.isChecked():
        if child_w.isVisible():
            self.toggle.setIcon(Expander.r_arrow)
            child_w.setVisible(False)
        else:
            self.toggle.setIcon(Expander.d_arrow)
            child_w.setVisible(True)


class TabWidget(ContainerBase):
    def __init__(self, tabpos='top', reorderable=False, detachable=False,
                 group=0):
        super(TabWidget, self).__init__()

        self.reorderable = reorderable
        self.detachable = detachable

        w = QtGui.QTabWidget()
        w.currentChanged.connect(self._cb_redirect)
        w.tabCloseRequested.connect(self._tab_close)
        w.setUsesScrollButtons(True)
        # w.setTabsClosable(True)
        if self.reorderable:
            w.setMovable(True)
        # w.tabInserted = self._tab_insert_cb
        # w.tabRemoved = self._tab_remove_cb
        self.widget = w
        self.set_tab_position(tabpos)

        for name in ('page-switch', 'page-close', 'page-move', 'page-detach'):
            self.enable_callback(name)

    def set_tab_position(self, tabpos):
        w = self.widget
        if tabpos == 'top':
            w.setTabPosition(QtGui.QTabWidget.North)
        elif tabpos == 'bottom':
            w.setTabPosition(QtGui.QTabWidget.South)
        elif tabpos == 'left':
            w.setTabPosition(QtGui.QTabWidget.West)
        elif tabpos == 'right':
            w.setTabPosition(QtGui.QTabWidget.East)

    def _cb_redirect(self, index):
        # get new index, because passed index can be out of date
        index = self.get_index()
        child = self.index_to_widget(index)
        if child is not None:
            self.make_callback('page-switch', child)

    def _tab_close(self, index):
        child = self.index_to_widget(index)
        self.make_callback('page-close', child)

    def add_widget(self, child, title=''):
        self.add_ref(child)
        child_w = child.get_widget()
        self.widget.addTab(child_w, title)
        # attach title to child
        child.extdata.tab_title = title
        self.make_callback('widget-added', child)

    def _remove(self, nchild, delete=False):
        idx = self.widget.indexOf(nchild)
        self.widget.removeTab(idx)

        nchild.setParent(None)
        if delete:
            nchild.deleteLater()

    def get_index(self):
        return self.widget.currentIndex()

    def set_index(self, idx):
        self.widget.setCurrentIndex(idx)
        # child = self.index_to_widget(idx)
        # child.focus()

    def index_of(self, child):
        return self.widget.indexOf(child.get_widget())

    def index_to_widget(self, idx):
        """Returns child corresponding to `idx`"""
        nchild = self.widget.widget(idx)
        if nchild is None:
            return nchild
        return self._native_to_child(nchild)

    def highlight_tab(self, idx, tf):
        tabbar = self.widget.tabBar()
        if not tf:
            color = QtHelp.QColor('black')
        else:
            color = QtHelp.QColor('green')
        tabbar.setTabTextColor(idx, color)


class StackWidget(ContainerBase):
    def __init__(self):
        super(StackWidget, self).__init__()

        self.widget = QtGui.QStackedWidget()

        for name in ['page-switch']:
            self.enable_callback(name)

    def add_widget(self, child, title=''):
        self.add_ref(child)
        child_w = child.get_widget()
        self.widget.addWidget(child_w)
        # attach title to child
        child.extdata.tab_title = title
        self.make_callback('widget-added', child)

    def get_index(self):
        return self.widget.currentIndex()

    def set_index(self, idx):
        _idx = self.widget.currentIndex()
        self.widget.setCurrentIndex(idx)

        child = self.index_to_widget(idx)
        if _idx != idx:
            self.make_callback('page-switch', child)
        # child.focus()

    def index_of(self, child):
        return self.widget.indexOf(child.get_widget())

    def index_to_widget(self, idx):
        nchild = self.widget.widget(idx)
        return self._native_to_child(nchild)


class MDIWidget(ContainerBase):
    def __init__(self, tabpos='top', mode='mdi'):
        super(MDIWidget, self).__init__()

        w = QtGui.QMdiArea()
        w.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        w.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        # See note below in add_widget()
        #w.subWindowActivated.connect(self._cb_redirect)

        # w.setSizePolicy(QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding,
        #                                   QtGui.QSizePolicy.Expanding))
        w.setTabsClosable(True)
        w.setTabsMovable(False)
        self.widget = w
        self.true_mdi = True
        self.cur_index = -1

        for name in ('page-switch', 'page-close'):
            self.enable_callback(name)

        self.set_tab_position(tabpos)
        self.set_mode(mode)

    def set_tab_position(self, tabpos):
        w = self.widget
        if tabpos == 'top':
            w.setTabPosition(QtGui.QTabWidget.North)
        elif tabpos == 'bottom':
            w.setTabPosition(QtGui.QTabWidget.South)
        elif tabpos == 'left':
            w.setTabPosition(QtGui.QTabWidget.West)
        elif tabpos == 'right':
            w.setTabPosition(QtGui.QTabWidget.East)

    def get_mode(self):
        if self.widget.viewMode() == QtGui.QMdiArea.TabbedView:
            return 'tabs'
        return 'mdi'

    def set_mode(self, mode):
        mode = mode.lower()
        if mode == 'tabs':
            self.widget.setViewMode(QtGui.QMdiArea.TabbedView)
        elif mode == 'mdi':
            self.widget.setViewMode(QtGui.QMdiArea.SubWindowView)
        else:
            raise ValueError("Don't understand mode='%s'" % (mode))

    def _cb_redirect(self, subwin):
        if subwin is not None:
            nchild = subwin.widget()
            child = self._native_to_child(nchild)
            try:
                self.cur_index = self.children.index(child)
            except Exception:
                self.cur_index = -1
            self.make_callback('page-switch', child)

    def _window_resized(self, event, subwin, widget):
        qsize = event.size()
        wd, ht = qsize.width(), qsize.height()
        # save size
        widget.extdata.mdi_size = (wd, ht)
        subwin._resizeEvent(event)

    def _window_moved(self, event, subwin, widget):
        qpos = event.pos()
        x, y = qpos.x(), qpos.y()
        # save position
        widget.extdata.mdi_pos = (x, y)
        subwin._moveEvent(event)

    def _window_closed(self, event, subwin, widget):
        nchild = subwin.widget()
        child = self._native_to_child(nchild)

        # let the application deal with this if desired in page-close
        # callback
        event.ignore()
        # self.widget.removeSubWindow(subwin)

        self.make_callback('page-close', child)

    def add_widget(self, child, title=''):
        self.add_ref(child)
        child_w = child.get_widget()
        subwin = QtGui.QMdiSubWindow(self.widget)
        # NOTE: we fire the page-switch callback by intercepting the
        # focus event on the subwindow, rather than off of the
        # subWindowActivated signal because the latter fires if
        # the widget accepts focus when the mouse enters the window,
        # whereas this approach one actually has to click in the window
        # or title bar.

        def _focus_cb(event):
            if event.gotFocus():
                self._cb_redirect(subwin)

        subwin.focusInEvent = _focus_cb
        # remove Qt logo from subwindow
        subwin.setWindowIcon(QIcon(QPixmap(1, 1)))
        subwin.setWidget(child_w)
        # attach title to child
        child.extdata.tab_title = title

        w = self.widget.addSubWindow(subwin)
        w._closeEvent = w.closeEvent
        w.closeEvent = lambda event: self._window_closed(event, w, child)

        # does child have a previously saved size
        size = child.extdata.get('mdi_size', None)
        if size is not None:
            wd, ht = size
            w.resize(wd, ht)

        # does child have a previously saved position
        pos = child.extdata.get('mdi_pos', None)
        if pos is not None:
            x, y = pos
            w.move(x, y)

        # Monkey-patching the widget to take control of resize and move
        # events
        w._resizeEvent = w.resizeEvent
        w.resizeEvent = lambda event: self._window_resized(event, w, child)
        w._moveEvent = w.moveEvent
        w.moveEvent = lambda event: self._window_moved(event, w, child)
        w.setWindowTitle(title)
        child_w.show()
        w.show()
        self.make_callback('widget-added', child)

    def _remove(self, nchild, delete=False):
        subwins = list(self.widget.subWindowList())
        l = [sw.widget() for sw in subwins]
        try:
            idx = l.index(nchild)
            subwin = subwins[idx]
        except (IndexError, ValueError) as e:
            subwin = None

        if subwin is not None:
            self.widget.removeSubWindow(subwin)
            subwin.deleteLater()

        nchild.setParent(None)
        if delete:
            nchild.deleteLater()

    def get_index(self):
        subwin = self.widget.activeSubWindow()
        if subwin is not None:
            return self._get_native_index(subwin.widget())
        return self.cur_index

    def _get_subwin(self, widget):
        for subwin in list(self.widget.subWindowList()):
            if subwin.widget() == widget:
                return subwin
        return None

    def set_index(self, idx):
        if 0 <= idx < len(self.children):
            child = self.children[idx]
            subwin = self._get_subwin(child.widget)
            if subwin is not None:
                self.widget.setActiveSubWindow(subwin)

    def index_of(self, child):
        nchild = child.get_widget()
        return self._get_native_index(nchild)

    def index_to_widget(self, idx):
        if 0 <= idx < len(self.children):
            return self.children[idx]
        return None

    def tile_panes(self):
        self.widget.tileSubWindows()

    def cascade_panes(self):
        self.widget.cascadeSubWindows()

    def use_tabs(self, tf):
        if tf:
            self.widget.setViewMode(QtGui.QMdiArea.TabbedView)
        else:
            self.widget.setViewMode(QtGui.QMdiArea.SubWindowView)


class ScrollArea(ContainerBase):
    def __init__(self):
        super(ScrollArea, self).__init__()

        self.widget = QtGui.QScrollArea()
        self.widget.setWidgetResizable(True)
        self.widget._resizeEvent = self.widget.resizeEvent
        self.widget.resizeEvent = self._resize_cb

        self.enable_callback('configure')

    def _resize_cb(self, event):
        self.widget._resizeEvent(event)

        rect = self.widget.geometry()
        x1, y1, x2, y2 = rect.getCoords()
        width = x2 - x1
        height = y2 - y1
        self.make_callback('configure', width, height)

    def set_widget(self, child):
        self.add_ref(child)
        self.widget.setWidget(child.get_widget())

    def scroll_to_end(self, vertical=True, horizontal=False):
        area = self.widget
        if vertical:
            area.verticalScrollBar().setValue(
                area.verticalScrollBar().maximum())
        if horizontal:
            area.horizontalScrollBar().setValue(
                area.horizontalScrollBar().maximum())


class Splitter(ContainerBase):
    def __init__(self, orientation='horizontal'):
        super(Splitter, self).__init__()

        w = QtGui.QSplitter()
        self.orientation = orientation
        if orientation == 'horizontal':
            w.setOrientation(QtCore.Qt.Horizontal)
        else:
            w.setOrientation(QtCore.Qt.Vertical)
        self.widget = w
        w.setChildrenCollapsible(True)

    def add_widget(self, child):
        self.add_ref(child)
        child_w = child.get_widget()
        self.widget.addWidget(child_w)
        self.make_callback('widget-added', child)

    def get_sizes(self):
        return list(self.widget.sizes())

    def set_sizes(self, sizes):
        return self.widget.setSizes(sizes)


class GridBox(ContainerBase):
    def __init__(self, rows=1, columns=1):
        super(GridBox, self).__init__()

        w = QtGui.QWidget()
        layout = QtGui.QGridLayout()
        w.setLayout(layout)
        self.widget = w

    def resize_grid(self, rows, columns):
        pass

    def set_row_spacing(self, val):
        self.widget.layout().setVerticalSpacing(val)

    def set_column_spacing(self, val):
        self.widget.layout().setHorizontalSpacing(val)

    def set_spacing(self, val):
        self.set_row_spacing(val)
        self.set_column_spacing(val)

    def add_widget(self, child, row, col, stretch=0):
        self.add_ref(child)
        w = child.get_widget()
        self.widget.layout().addWidget(w, row, col)
        self.make_callback('widget-added', child)


class ToolbarAction(WidgetBase):
    def __init__(self):
        super(ToolbarAction, self).__init__()

        self.widget = None
        self.enable_callback('activated')

    def _cb_redirect(self, *args):
        if self.widget.isCheckable():
            tf = self.widget.isChecked()
            self.make_callback('activated', tf)
        else:
            self.make_callback('activated')

    def set_state(self, tf):
        self.widget.setChecked(tf)

    def get_state(self):
        return self.widget.isChecked()


class Toolbar(ContainerBase):
    def __init__(self, orientation='horizontal'):
        super(Toolbar, self).__init__()

        w = QtGui.QToolBar()
        if orientation == 'horizontal':
            w.setOrientation(QtCore.Qt.Horizontal)
        else:
            w.setOrientation(QtCore.Qt.Vertical)
        self.widget = w
        self._menu = None
        self.widget.setStyleSheet(
            """
            QToolBar { padding: 0; spacing: 0; }\n
            QToolBar QToolButton { padding: 0; margin: 0; }\n
            """)

    def add_action(self, text, toggle=False, iconpath=None, iconsize=None):
        child = ToolbarAction()
        if iconpath:
            image = QImage(iconpath)
            wd, ht = 24, 24
            if iconsize is not None:
                wd, ht = iconsize
            qsize = QtCore.QSize(wd, ht)
            image = image.scaled(qsize)
            pixmap = QPixmap.fromImage(image)
            iconw = QIcon(pixmap)
            action = self.widget.addAction(iconw, text,
                                           child._cb_redirect)
            if text is not None and len(text) > 0:
                self.widget.setToolButtonStyle(QtCore.Qt.ToolButtonTextUnderIcon)
        else:
            action = self.widget.addAction(text, child._cb_redirect)
        action.setCheckable(toggle)
        child.widget = action
        self.add_ref(child)
        return child

    def add_widget(self, child):
        self.add_ref(child)
        w = child.get_widget()
        # in toolbars, generally don't want widgets to take up any more
        # space than necessary
        w.setSizePolicy(QtGui.QSizePolicy.Fixed,
                        QtGui.QSizePolicy.Fixed)
        self.widget.addWidget(w)
        self.make_callback('widget-added', child)

    def add_menu(self, text, menu=None, mtype='tool'):
        if menu is None:
            menu = Menu()
        if mtype == 'tool':
            child = self.add_action(text)
        else:
            child = Label(text, style='clickable', menu=menu)
            self.add_widget(child)
            child.add_callback('released', lambda w: menu.hide())

        child.add_callback('activated', lambda w: menu.popup())
        return menu

    def add_separator(self):
        self.widget.addSeparator()


class MenuAction(WidgetBase):
    def __init__(self, text=None, checkable=False):
        super(MenuAction, self).__init__()

        self.widget = None
        self.text = text
        self.checkable = checkable
        self.enable_callback('activated')

    def set_state(self, tf):
        if not self.checkable:
            raise ValueError("Not a checkable menu item")
        self.widget.setChecked(tf)

    def get_state(self):
        return self.widget.isChecked()

    def _cb_redirect(self, *args):
        if self.widget.isCheckable():
            tf = self.widget.isChecked()
            self.make_callback('activated', tf)
        else:
            self.make_callback('activated')


class Menu(ContainerBase):
    def __init__(self):
        super(Menu, self).__init__()

        # NOTE: this get's overwritten if created from Menubar
        self.widget = QtGui.QMenu()
        self.menus = Bunch.Bunch(caseless=True)

    def add_widget(self, child):
        w = self.widget.addAction(child.text, lambda: child._cb_redirect())
        if child.checkable:
            w.setCheckable(True)
        child.widget = w
        self.add_ref(child)
        self.make_callback('widget-added', child)

    def add_name(self, name, checkable=False):
        child = MenuAction(text=name, checkable=checkable)
        self.add_widget(child)
        return child

    def add_menu(self, name):
        menu_w = self.widget.addMenu(name)
        child = Menu()
        child.widget = menu_w
        self.add_ref(child)
        self.menus[name] = child
        return child

    def add_separator(self):
        self.widget.addSeparator()

    def popup(self, widget=None):
        if widget is not None:
            w = widget.get_widget()
            if w.isEnabled():
                # self.widget.popup(w.mapToGlobal(QtCore.QPoint(0, 0)))
                self.widget.exec_(w.mapToGlobal(QtCore.QPoint(0, 0)))
        else:
            if self.widget.isEnabled():
                self.widget.exec_(QCursor.pos())

    def get_menu(self, name):
        return self.menus[name]


class Menubar(ContainerBase):
    def __init__(self):
        super(Menubar, self).__init__()

        self.widget = QtGui.QMenuBar()
        if hasattr(self.widget, 'setNativeMenuBar'):
            self.widget.setNativeMenuBar(False)
        self.menus = Bunch.Bunch(caseless=True)

    def add_widget(self, child, name):
        if not isinstance(child, Menu):
            raise ValueError("child widget needs to be a Menu object")
        menu_w = self.widget.addMenu(name)
        child.widget = menu_w
        self.add_ref(child)
        self.menus[name] = child
        self.make_callback('widget-added', child)
        return child

    def add_name(self, name):
        menu_w = self.widget.addMenu(name)
        child = Menu()
        child.widget = menu_w
        self.add_ref(child)
        self.menus[name] = child
        return child

    def get_menu(self, name):
        return self.menus[name]


class TopLevelMixin(object):

    def __init__(self, title=None):

        self.widget.closeEvent = lambda event: self._quit(event)
        self.widget.destroyed = self._destroyed_cb

        if title is not None:
            self.widget.setWindowTitle(title)

        self.enable_callback('close')

    def _quit(self, event):
        # event.accept()
        # let application decide how to handle this
        event.ignore()
        self.close()

    def _closeEvent(self, *args):
        self.close()

    def close(self):
        # self.widget.deleteLater()
        # self.widget = None
        self.make_callback('close')

    def _destroyed_cb(self, event, *args):
        event.accept()

    def raise_(self):
        self.widget.raise_()
        self.widget.activateWindow()

    def lower(self):
        self.widget.lower()

    def focus(self):
        self.widget.raise_()
        self.widget.activateWindow()

    def move(self, x, y):
        self.widget.move(x, y)

    def maximize(self):
        self.widget.showMaximized()

    def unmaximize(self):
        self.widget.showNormal()

    def fullscreen(self):
        self.widget.showFullScreen()

    def unfullscreen(self):
        self.widget.showNormal()

    def is_fullscreen(self):
        return self.widget.isFullScreen()

    def iconify(self):
        self.hide()

    def uniconify(self):
        self.widget.showNormal()

    def set_title(self, title):
        self.widget.setWindowTitle(title)


class TopLevel(TopLevelMixin, ContainerBase):

    def __init__(self, title=None):
        ContainerBase.__init__(self)

        widget = QtHelp.TopLevel()
        self.widget = widget
        box = QtGui.QVBoxLayout()
        box.setContentsMargins(0, 0, 0, 0)
        box.setSpacing(0)
        widget.setLayout(box)

        TopLevelMixin.__init__(self, title=title)

    def set_widget(self, child):
        self.add_ref(child)
        child_w = child.get_widget()
        self.widget.layout().addWidget(child_w)


class Application(Callback.Callbacks):

    def __init__(self, logger=None):
        global _app
        super(Application, self).__init__()

        self.logger = logger
        self.window_list = []

        self.window_dict = {}
        self.wincnt = 0

        if have_pyqt4:
            QtGui.QApplication.setGraphicsSystem('raster')
        app = QtGui.QApplication([])
        # app.lastWindowClosed.connect(lambda *args: self._quit())
        self._qtapp = app
        _app = self

        # Get screen size
        desktop = self._qtapp.desktop()
        rect = desktop.availableGeometry()
        size = rect.size()
        self.screen_wd = size.width()
        self.screen_ht = size.height()

        # Get screen resolution
        xdpi = desktop.physicalDpiX()
        ydpi = desktop.physicalDpiY()
        self.screen_res = max(xdpi, ydpi)

        for name in ('shutdown', ):
            self.enable_callback(name)

    def get_screen_size(self):
        return (self.screen_wd, self.screen_ht)

    def process_events(self):
        self._qtapp.processEvents()

    def process_end(self):
        self._qtapp.quit()

    def add_window(self, window, wid=None):
        if wid is None:
            wid = 'win%d' % (self.wincnt)
            self.wincnt += 1
        window.wid = wid
        window.url = ''
        window.app = self

        self.window_dict[wid] = window

    def get_window(self, wid):
        return self.window_dict[wid]

    def has_window(self, wid):
        return wid in self.window_dict

    def get_wids(self):
        return list(self.window_dict.keys())

    def make_window(self, title=None):
        w = TopLevel(title=title)
        self.add_window(w)
        return w

    def make_timer(self):
        return QtHelp.Timer()

    def mainloop(self):
        self._qtapp.exec_()

    def quit(self):
        self._qtapp.quit()


class Dialog(TopLevelMixin, WidgetBase):

    def __init__(self, title='', flags=None, buttons=[],
                 parent=None, modal=False):
        WidgetBase.__init__(self)

        if parent is not None:
            parent = parent.get_widget()
        self.widget = QtGui.QDialog(parent)
        self.widget.setModal(modal)

        vbox = QtGui.QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        self.widget.setLayout(vbox)

        TopLevelMixin.__init__(self, title=title)

        self.content = VBox()
        vbox.addWidget(self.content.get_widget(), stretch=1)

        if len(buttons) > 0:
            hbox_w = QtGui.QWidget()
            hbox = QtGui.QHBoxLayout()
            hbox_w.setLayout(hbox)

            for name, val in buttons:
                btn = QtGui.QPushButton(name)

                def cb(val):
                    return lambda: self._cb_redirect(val)

                btn.clicked.connect(cb(val))
                hbox.addWidget(btn, stretch=0)

            vbox.addWidget(hbox_w, stretch=0)
            # self.widget.closeEvent = lambda event: self.delete()

        self.enable_callback('activated')

    def _cb_redirect(self, val):
        self.make_callback('activated', val)

    def get_content_area(self):
        return self.content


class SaveDialog(QtGui.QFileDialog):

    def __init__(self, title=None, selectedfilter=None):
        super(SaveDialog, self).__init__()

        self.title = title
        self.selectedfilter = selectedfilter
        self.widget = self

    def get_path(self):
        res = self.getSaveFileName(self, self.title, '', self.selectedfilter)

        if isinstance(res, tuple):
            res = res[0]

        if (res and self.selectedfilter is not None and
                not res.endswith(self.selectedfilter[1:])):
            res += self.selectedfilter[1:]
        if res == '':
            # user cancelled dialog
            res = None

        return res


class DragPackage(object):
    def __init__(self, src_widget):
        self.src_widget = src_widget
        self._drag = QtHelp.QDrag(self.src_widget)
        self._data = QtCore.QMimeData()
        self._drag.setMimeData(self._data)

    def set_urls(self, urls):
        _urls = [QtCore.QUrl(url) for url in urls]
        self._data.setUrls(_urls)

    def set_text(self, text):
        self._data.setText(text)

    def start_drag(self):
        if QtHelp.have_pyqt5:
            self._drag.exec_(QtCore.Qt.MoveAction)
        else:
            self._drag.start(QtCore.Qt.MoveAction)


# MODULE FUNCTIONS

def name_mangle(name, pfx=''):
    newname = []
    for c in name.lower():
        if not (c.isalpha() or c.isdigit() or (c == '_')):
            newname.append('_')
        else:
            newname.append(c)
    return pfx + ''.join(newname)


def hadjust(w, orientation):
    """Ostensibly, a function to reduce the vertical footprint of a widget
    that is normally used in a vertical stack (usually a Splitter), when it
    is instead used in a horizontal orientation.
    """
    if orientation != 'horizontal':
        return w
    # This currently does not seem to be needed for most plugins that are
    # coded to flow either vertically or horizontally and, in fact, reduces
    # the visual asthetic somewhat.
    ## spl = Splitter(orientation='vertical')
    ## spl.add_widget(w)
    ## spl.add_widget(Label(''))
    ## return spl
    return w


def make_widget(title, wtype):
    if wtype == 'label':
        w = Label(title)
        w.widget.setAlignment(QtCore.Qt.AlignRight)
    elif wtype == 'llabel':
        w = Label(title)
        w.widget.setAlignment(QtCore.Qt.AlignLeft)
    elif wtype == 'entry':
        w = TextEntry()
        # w.widget.setMaxLength(12)
    elif wtype == 'entryset':
        w = TextEntrySet()
        # w.widget.setMaxLength(12)
    elif wtype == 'combobox':
        w = ComboBox()
    elif wtype == 'spinbutton':
        w = SpinBox(dtype=int)
    elif wtype == 'spinfloat':
        w = SpinBox(dtype=float)
    elif wtype == 'vbox':
        w = VBox()
    elif wtype == 'hbox':
        w = HBox()
    elif wtype == 'hscale':
        w = Slider(orientation='horizontal')
    elif wtype == 'vscale':
        w = Slider(orientation='vertical')
    elif wtype == 'checkbutton':
        w = CheckBox(title)
    elif wtype == 'radiobutton':
        w = RadioButton(title)
    elif wtype == 'togglebutton':
        w = ToggleButton(title)
    elif wtype == 'button':
        w = Button(title)
    elif wtype == 'spacer':
        w = Label('')
    elif wtype == 'textarea':
        w = TextArea(editable=True)
    elif wtype == 'toolbar':
        w = Toolbar()
    elif wtype == 'progress':
        w = ProgressBar()
    elif wtype == 'menubar':
        w = Menubar()
    else:
        raise ValueError("Bad wtype=%s" % wtype)
    return w


def build_info(captions, orientation='vertical'):
    # numrows = len(captions)
    numcols = reduce(lambda acc, tup: max(acc, len(tup)), captions, 0)
    if (numcols % 2) != 0:
        raise ValueError("Column spec is not an even number")
    numcols = int(numcols // 2)

    widget = QtGui.QWidget()
    table = QtGui.QGridLayout()
    widget.setLayout(table)
    table.setVerticalSpacing(2)
    table.setHorizontalSpacing(4)
    table.setContentsMargins(2, 2, 2, 2)

    wb = Bunch.Bunch()
    row = 0
    for tup in captions:
        col = 0
        while col < numcols:
            idx = col * 2
            if idx < len(tup):
                title, wtype = tup[idx:idx + 2]
                if not title.endswith(':'):
                    name = name_mangle(title)
                else:
                    name = name_mangle('lbl_' + title[:-1])
                w = make_widget(title, wtype)
                table.addWidget(w.widget, row, col)
                wb[name] = w
            col += 1
        row += 1

    w = wrap(widget)
    w = hadjust(w, orientation=orientation)

    return w, wb


def wrap(native_widget):
    wrapper = WidgetBase()
    wrapper.widget = native_widget
    return wrapper


# END



#
# layers.py -- classes for special compound objects layered on
#                   ginga canvases.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from ginga.canvas.CanvasObject import (CanvasObjectBase,
                                       register_canvas_types)
from ginga import Mixins
from ginga.misc.log import get_logger

from ..CompoundMixin import CompoundMixin
from ..CanvasMixin import CanvasMixin
from ..DrawingMixin import DrawingMixin

__all__ = ['CompoundObject', 'Canvas', 'DrawingCanvas']


class CompoundObject(CompoundMixin, CanvasObjectBase):
    """Compound object on a ImageViewCanvas.
    Parameters are the child objects making up the compound object.
    Objects are drawn in the order listed. Example:

    .. code-block: Python

        CompoundObject(Point(x, y, radius, ...),
                       Circle(x, y, radius, ...))

    This makes a point inside a circle.
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            ## Param(name='coord', type=str, default='data',
            ##       valid=['data', 'wcs'],
            ##       description="Set type of coordinates"),
        ]

    def __init__(self, *objects, **kwdargs):
        CanvasObjectBase.__init__(self, **kwdargs)
        CompoundMixin.__init__(self)
        self.objects = list(objects)
        self.logger = get_logger('foo', log_stderr=True, level=10)

        self.kind = 'compound'
        self.editable = False


class Canvas(CanvasMixin, CompoundObject):
    """Class to handle canvas in Ginga."""
    @classmethod
    def get_params_metadata(cls):
        return [
            ## Param(name='coord', type=str, default='data',
            ##       valid=['data', 'wcs'],
            ##       description="Set type of coordinates"),
        ]

    def __init__(self, *objects, **kwdargs):
        CompoundObject.__init__(self, *objects, **kwdargs)
        CanvasMixin.__init__(self)

        self.kind = 'canvas'
        self.editable = False


class DrawingCanvas(Mixins.UIMixin, DrawingMixin, Canvas):
    """Drawing canvas."""
    def __init__(self, **kwdargs):
        Canvas.__init__(self, **kwdargs)
        DrawingMixin.__init__(self)
        Mixins.UIMixin.__init__(self)

        self.kind = 'drawingcanvas'
        self.editable = False


class ConstructedCanvas(DrawingMixin, Canvas):
    """Constructed canvas from a list of specifications.

    Parameters are specifications of child objects, where each specification
    is a map specifying 'type' and (optionally) 'args' and 'kwargs'.
    If present, 'args' is a sequence of arguments and 'kwargs' is a map
    of keyword arguments to provide to the constructor of the child.

    Example:

    .. code-block: Python

        ConstructedCanvas([dict(type='point', args=(x, y, radius),
                                kwargs=dict(color='red')),
                           dict(type='circle', args=(x, y, radius),
                                kwargs=dict(color='yellow'))])

    This makes a point inside a circle.
    """
    def __init__(self, spec_list, **kwdargs):
        Canvas.__init__(self, **kwdargs)
        DrawingMixin.__init__(self)
        self.objects = self.build_objects(spec_list)

        self.kind = 'constructed'
        self.editable = False

    def build_objects(self, spec_list):
        return [self.build_object(spec) for spec in spec_list]

    def build_object(self, spec):
        ctype = spec.get('type', None)
        if ctype is None:
            raise ValueError("Item specification needs a 'type' designator: %s" % (
                str(spec)))

        # TODO: we need to be a subclass of DrawingMixin in order to get
        # access to the get_draw_class() method.  Otherwise we could just
        # be a subclass of CompoundObject. See if this can be fixed.
        draw_class = self.get_draw_class(ctype)

        args = spec.get('args', [])
        kwargs = spec.get('kwargs', {})

        if isinstance(draw_class, CompoundObject):
            # special case for compound objects: need to have actual objects
            # in constructor args, not specifications
            args = self.build_objects(args)

        return draw_class(*args, **kwargs)


catalog = dict(compoundobject=CompoundObject, canvas=Canvas,
               drawingcanvas=DrawingCanvas,
               constructedcanvas=ConstructedCanvas)
register_canvas_types(catalog)

# END



#
# Settings.py -- Simple class to manage stateful user preferences.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import os
import re
import ast

import numpy as np

from . import Callback
from . import Bunch

unset_value = "^^UNSET^^"
regex_array = re.compile(r'^\s*array\((\[.*\])\s*(,\s*dtype=(.+)\s*)?\)\s*$',
                         flags=re.DOTALL)


class SettingError(Exception):
    pass


class Setting(Callback.Callbacks):

    def __init__(self, value=unset_value, name=None, logger=None,
                 check_fn=None):
        Callback.Callbacks.__init__(self)

        self.value = value
        self._unset = np.isscalar(value) and value == unset_value
        self.name = name
        self.logger = logger
        if check_fn is None:
            check_fn = self._check_none
        self.check_fn = check_fn

        # For callbacks
        for name in ('set', ):
            self.enable_callback(name)

    def _check_none(self, value):
        return value

    def set(self, value, callback=True):
        self.value = self.check_fn(value)
        if callback:
            self.make_callback('set', value)

    def get(self, *args):
        if self._unset:
            if len(args) == 0:
                raise KeyError("setting['%s'] value is not set!" % (
                    self.name))
            else:
                if len(args) != 1:
                    raise SettingError("Illegal parameter use to get(): %s" % (
                        str(args)))
                return args[0]
        return self.value

    def __repr__(self):
        return repr(self.value)

    def __str__(self):
        return str(self.value)


class SettingGroup(object):

    def __init__(self, name=None, logger=None, preffile=None):
        self.name = name
        self.logger = logger
        self.preffile = preffile

        # TODO: add group change callback?
        self.group = Bunch.Bunch()
        self._group_b = Bunch.Bunch()

    def add_settings(self, **kwdargs):
        for key, value in kwdargs.items():
            if key not in self.group:
                setting = Setting(value=value, name=key,
                                  logger=self.logger)
                self._mk_add_callback(key, setting)
                self.group[key] = setting
            # add to backup group
            if key not in self._group_b:
                self._group_b[key] = setting

    def get_setting(self, key):
        return self.group[key]

    def keys(self):
        return self.group.keys()

    def _mk_add_callback(self, key, setting):
        setting._add_callback = setting.add_callback

        def _add_callback(*args, **kwargs):
            self.group[key]._add_callback(*args, **kwargs)
            self._group_b[key]._add_callback(*args, **kwargs)

        setting.add_callback = _add_callback

    def share_settings(self, other, keylist=None, include_callbacks=True,
                       callback=True):
        """Sharing settings with `other`
        """
        if keylist is None:
            keylist = self.group.keys()
        if include_callbacks:
            for key in keylist:
                oset, mset = other.group[key], self.group[key]
                other.group[key] = mset
                oset.merge_callbacks_to(mset)
        if callback:
            # make callbacks only after all items are set in the group
            # TODO: only make callbacks for values that changed?
            for key in keylist:
                other.group[key].make_callback('set', other.group[key].value)

    def unshare_settings(self, keylist=None, callback=True):
        if keylist is None:
            keylist = self._group_b.keys()
        for key in keylist:
            # copy value from current setting, while restoring old setting
            if self.group[key] is not self._group_b[key]:
                value = self.group[key].value
                self.group[key] = self._group_b[key]
                self.group[key].set(value, callback=False)
        if callback:
            # make callbacks only after all items are set in the group
            # TODO: only make callbacks for values that changed?
            for key in keylist:
                self.group[key].make_callback('set', self.group[key].value)

    def copy_settings(self, other, keylist=None, include_callbacks=False,
                      callback=True):
        if keylist is None:
            keylist = self.group.keys()
        d = {key: self.get(key) for key in keylist}

        if include_callbacks:
            for key in keylist:
                oset, mset = other.group[key], self.group[key]
                mset.merge_callbacks_to(oset)

        other.set_dict(d, callback=callback)

    def setdefault(self, key, value):
        if key in self.group:
            return self.group[key].get(value)
        else:
            d = {key: value}
            self.add_settings(**d)
            return self.group[key].get(value)

    def add_defaults(self, **kwdargs):
        for key, value in kwdargs.items():
            self.setdefault(key, value)

    def set_defaults(self, **kwdargs):
        return self.add_defaults(**kwdargs)

    def get(self, *args):
        key = args[0]
        if len(args) == 1:
            return self.group[key].get()
        else:
            if key in self.group:
                return self.group[key].get(*args[1:])
            if len(args) > 2:
                raise SettingError("Illegal parameter use to get(): %s" % (
                    str(args)))
            return args[1]

    def get_dict(self, keylist=None):
        if keylist is None:
            keylist = self.group.keys()
        return dict([[name, self.group[name].value]
                     for name in keylist])

    def set_dict(self, d, callback=True):
        for key, value in d.items():
            if key not in self.group:
                self.setdefault(key, value)
            else:
                self.group[key].set(value, callback=False)

        if callback:
            # make callbacks only after all items are set in the group
            for key, value in d.items():
                self.group[key].make_callback('set', value)

    def set(self, callback=True, **kwdargs):
        self.set_dict(kwdargs, callback=callback)

    def __getitem__(self, key):
        return self.group[key].value

    def __setitem__(self, key, value):
        self.group[key].set(value)

    # TODO: Should deprecate this and encourage __contains__ like Python dict
    def has_key(self, key):
        return key in self.group

    def __contains__(self, key):
        return key in self.group

    def clear_callbacks(self, keylist=None):
        if keylist is None:
            keylist = self.group.keys()

        for key in keylist:
            self.group[key].clear_callback('set')

    def load(self, onError='raise', buf=None):
        try:
            d = {}
            if buf is None:
                with open(self.preffile, 'r') as in_f:
                    buf = in_f.read()
            for line_no, line in enumerate(buf.split('\n')):
                line = line.strip()
                # skip comments and anything that doesn't look like an
                # assignment
                if line.startswith('#') or (not ('=' in line)):
                    continue
                else:
                    try:
                        i = line.index('=')
                        key = line[:i].strip()
                        val_s = line[i + 1:].strip()
                        match = regex_array.match(val_s)
                        if match:
                            data, _x, dtype = match.groups()
                            # special case for parsing numpy arrays
                            val = np.asarray(ast.literal_eval(data),
                                             dtype=dtype)
                        else:
                            val = ast.literal_eval(val_s)
                        d[key] = val
                    except Exception as e:
                        if self.logger is not None:
                            self.logger.warn("Error loading '%s' at line %d: "
                                             "%s" % (self.preffile, line_no,
                                                     str(e)))
                        continue

            self.set_dict(d)
        except Exception as e:
            errmsg = "Error opening settings file (%s): %s" % (
                self.preffile, str(e))
            if onError == 'silent':
                pass
            elif onError == 'warn':
                self.logger.warning(errmsg)
            else:
                raise SettingError(errmsg)

    def _check(self, d):
        if isinstance(d, dict):
            for key, value in d.items():
                d[key] = self._check(value)
            return d
        try:
            if np.isnan(d):
                return 0.0
            elif np.isinf(d):
                return 0.0
        except Exception:
            pass
        return d

    def _save(self, out_f, keys, d):
        for key in keys:
            val_s = repr(d[key])
            match = regex_array.match(val_s)
            if match:
                # special processing for numpy arrays
                val_s = val_s.replace('\n', ' ')
            out_f.write("%s = %s\n" % (key, val_s))

    def save(self, keylist=None, output=None):
        d = self.get_dict(keylist=keylist)
        # sanitize data -- hard to parse NaN or Inf
        self._check(d)
        try:
            # sort keys for easy reading/editing
            keys = list(d.keys())
            keys.sort()

            if output is None:
                output = self.preffile
            if isinstance(output, str):
                with open(output, 'w') as out_f:
                    self._save(out_f, keys, d)
            else:
                self._save(output, keys, d)

        except Exception as e:
            errmsg = "Error writing settings output: %s" % (str(e))
            self.logger.error(errmsg)

    ########################################################
    ### NON-PEP8 PREDECESSORS: TO BE DEPRECATED

    addSettings = add_settings
    getSetting = get_setting
    shareSettings = share_settings
    copySettings = copy_settings
    addDefaults = add_defaults
    setDefaults = set_defaults
    getDict = get_dict
    setDict = set_dict


class Preferences(object):

    def __init__(self, basefolder=None, logger=None):
        self.folder = basefolder
        self.logger = logger
        self.settings = Bunch.Bunch(caseless=True)

    def set_defaults(self, category, **kwdargs):
        self.settings[category].add_defaults(**kwdargs)

    def get_settings(self, category):
        return self.settings[category]

    def remove_settings(self, category):
        del self.settings[category]

    def get_dict_category(self, category):
        return self.settings[category].get_dict()

    def create_category(self, category):
        if category not in self.settings:
            suffix = '.cfg'
            path = os.path.join(self.folder, category + suffix)
            self.settings[category] = SettingGroup(logger=self.logger,
                                                   name=category,
                                                   preffile=path)
        return self.settings[category]

    def get_base_folder(self):
        return self.folder

    def get_dict(self):
        return dict([[name, self.settings[name].get_dict()] for name in
                     self.settings.keys()])

    ########################################################
    ### NON-PEP8 PREDECESSORS: TO BE DEPRECATED

    setDefaults = set_defaults
    getSettings = get_settings
    createCategory = create_category
    get_baseFolder = get_base_folder
    getDict = get_dict


#END



#
# colorramp.py -- utility functions for making ginga colormaps
#


def hue_sat_to_cmap(hue, sat):
    """Mkae a color map from a hue and saturation value.
    """
    import colorsys

    # normalize to floats
    hue = float(hue) / 360.0
    sat = float(sat) / 100.0

    res = []
    for val in range(256):
        hsv_val = float(val) / 255.0
        r, g, b = colorsys.hsv_to_rgb(hue, sat, hsv_val)
        res.append((r, g, b))

    return res


def rgbarr_to_colormap(rgb_data_np):

    data = rgb_data_np

    ht, wd, dp = data.shape
    assert dp == 3

    mid_y = ht // 2

    slc = data[mid_y, :, :]
    l = [tuple(slc[int(float(i) / 256 * wd), :]) for i in range(0, 256)]
    assert len(l) == 256

    clst = [(r / 255., g / 255., b / 255.) for r, g, b in l]
    return clst


def png_to_colormap(png_file):

    from ginga.util import io_rgb
    img = io_rgb.load_file(png_file)
    data = img.get_data()

    return rgbarr_to_colormap(data)


def print_colorramp(name, clst):
    print("cmap_%s = (")

    for r, g, b in clst:
        print("                (%f, %f, %f)," % (r, g, b))

    print("                )")
    print("")

# END



#
# CanvasRenderBokeh.py -- for rendering into a Bokeh widget
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import numpy as np

# Bokeh imports
#from bokeh.plotting import figure

#from ginga.canvas.mixins import *

from ginga.canvas import render
# force registration of all canvas types
import ginga.canvas.types.all  # noqa
from ginga import trcalc

from . import BokehHelp


class RenderContext(render.RenderContextBase):

    def __init__(self, renderer, viewer, surface):
        render.RenderContextBase.__init__(self, renderer, viewer)
        self.shape = None

        # TODO: encapsulate this drawable
        self.cr = BokehHelp.BokehContext(self.viewer.figure)

        self.pen = None
        self.brush = None
        self.font = None

    def set_line_from_shape(self, shape):
        # TODO: support line width and style
        alpha = getattr(shape, 'alpha', 1.0)
        self.pen = self.cr.get_pen(shape.color, alpha=alpha)

    def set_fill_from_shape(self, shape):
        fill = getattr(shape, 'fill', False)
        if fill:
            if hasattr(shape, 'fillcolor') and shape.fillcolor:
                color = shape.fillcolor
            else:
                color = shape.color
            alpha = getattr(shape, 'alpha', 1.0)
            alpha = getattr(shape, 'fillalpha', alpha)
            self.brush = self.cr.get_brush(color, alpha=alpha)
        else:
            self.brush = None

    def set_font_from_shape(self, shape):
        if hasattr(shape, 'font'):
            if hasattr(shape, 'fontsize') and shape.fontsize is not None:
                fontsize = shape.fontsize
            else:
                fontsize = shape.scale_font(self.viewer)
            fontsize = self.scale_fontsize(fontsize)
            alpha = getattr(shape, 'alpha', 1.0)
            self.font = self.cr.get_font(shape.font, fontsize, shape.color,
                                         alpha=alpha)
        else:
            self.font = None

    def initialize_from_shape(self, shape, line=True, fill=True, font=True):
        if line:
            self.set_line_from_shape(shape)
        if fill:
            self.set_fill_from_shape(shape)
        if font:
            self.set_font_from_shape(shape)

    def set_line(self, color, alpha=1.0, linewidth=1, style='solid'):
        # TODO: support style
        self.pen = self.cr.get_pen(color, alpha=alpha, linewidth=linewidth,
                                   linestyle=style)

    def set_fill(self, color, alpha=1.0):
        if color is None:
            self.brush = None
        else:
            self.brush = self.cr.get_brush(color, alpha=alpha)

    def set_font(self, fontname, fontsize):
        fontsize = self.scale_fontsize(fontsize)
        self.font = self.cr.get_font(fontname, fontsize, 'black',
                                     alpha=1.0)

    def text_extents(self, text):
        return self.cr.text_extents(text, self.font)

    ##### DRAWING OPERATIONS #####

    def draw_text(self, cx, cy, text, rot_deg=0.0):
        self.cr.init(angle=[np.radians(rot_deg)])
        self.cr.update_font(self.pen, self.font)

        self.cr.plot.text(x=[cx], y=[cy], text=[text], **self.cr.kwdargs)

    def draw_polygon(self, cpoints):
        self.cr.init()
        self.cr.update_patch(self.pen, self.brush)

        xy = trcalc.strip_z(cpoints)

        self.cr.plot.patches(xs=[xy.T[0]], ys=[xy.T[1]], **self.cr.kwdargs)

    def draw_circle(self, cx, cy, cradius):
        self.cr.init()
        self.cr.update_patch(self.pen, self.brush)

        self.cr.plot.circle(x=[cx], y=[cy], radius=[cradius],
                            **self.cr.kwdargs)

    def draw_bezier_curve(self, verts):
        self.cr.init()
        self.cr.update_line(self.pen)

        cx, cy = verts.T[0], verts.T[1]

        self.cr.plot.bezier(x0=[cx[0]], y0=[cy[0]],
                            x1=[cx[3]], y1=[cy[3]],
                            cx0=[cx[1]], cy0=[cy[1]],
                            cx1=[cx[2]], cy1=[cy[2]],
                            **self.cr.kwdargs)

    def draw_ellipse(self, cx, cy, cxradius, cyradius, theta):
        self.cr.init()
        self.cr.update_patch(self.pen, self.brush)

        self.cr.plot.oval(x=[cx], y=[cy],
                          width=[cxradius * 2.0], height=[cyradius * 2.0],
                          angle=[np.radians(theta)], **self.cr.kwdargs)

    def draw_line(self, cx1, cy1, cx2, cy2):
        self.cr.init()
        self.cr.update_line(self.pen)

        self.cr.plot.line(x=[cx1, cx2], y=[cy1, cy2], **self.cr.kwdargs)

    def draw_path(self, cpoints):
        self.cr.init()

        self.cr.update_line(self.pen)

        xy = trcalc.strip_z(cpoints)

        self.cr.plot.line(x=xy.T[0], y=xy.T[1], **self.cr.kwdargs)


class CanvasRenderer(render.RendererBase):

    def __init__(self, viewer):
        render.RendererBase.__init__(self, viewer)

        self.kind = 'bokeh'
        self.rgb_order = 'RGBA'
        self.surface = None

    def setup_cr(self, shape):
        cr = RenderContext(self, self.viewer, self.surface)
        cr.initialize_from_shape(shape)
        return cr

    def get_dimensions(self, shape):
        cr = self.setup_cr(shape)
        cr.set_font_from_shape(shape)
        return cr.text_extents(shape.text)


#END



# Uses Voki Codder's solution
# http://vokicodder.blogspot.in/2011/02/numpy-arrays-to-video.html

import subprocess


class VideoSink(object):

    def __init__(self, size, filename="output", rate=2, byteorder="Y8"):
        self.size = size
        self.cmdstring = (
            'mencoder', '/dev/stdin', '-demuxer', 'rawvideo', '-rawvideo',
            'w=%i:h=%i' % size[::-1] + ":fps=%i:format=%s" % (rate, byteorder),
            '-o', filename, '-ovc', 'lavc')

    def open(self):
        self.p = subprocess.Popen(self.cmdstring, stdin=subprocess.PIPE,
                                  shell=False)

    def write(self, image):
        assert image.shape == self.size
        self.p.stdin.write(image.tostring())

    def close(self):
        self.p.stdin.close()



"""

Usage:
  $ bokeh serve example1.py

you will see something like this in the output:
2017-11-29 17:00:48,368 Starting Bokeh server version 0.12.7 (running on Tornado 4.3)
2017-11-29 17:00:48,371 Bokeh app running at: http://localhost:5006/example1
2017-11-29 17:00:48,371 Starting Bokeh server with process id: 14214

Visit the URL with a browser to interact with the GUI.  Enter a valid
FITS file path into the box labeled "File:" and press Enter.  Use the
slider to zoom the image.
"""

import sys

from bokeh.io import curdoc
from bokeh.layouts import column
from bokeh.plotting import figure
#from bokeh.client import push_session
from bokeh.models import TextInput, Slider

from ginga.web.bokehw import ImageViewBokeh as ib
from ginga.misc import log
from ginga.util.loader import load_data


def main(options, args):

    #logger = log.get_logger("ginga", options=options)
    logger = log.get_logger("ginga", level=20, log_file="/tmp/ginga.log")

    #TOOLS = "pan,wheel_zoom,box_select,tap"
    TOOLS = "box_select"

    # create a new plot with default tools, using figure
    fig = figure(x_range=[0, 600], y_range=[0, 600],
                 plot_width=600, plot_height=600, tools=TOOLS)

    viewer = ib.CanvasView(logger)
    viewer.set_figure(fig)

    bd = viewer.get_bindings()
    bd.enable_all(True)

    ## box_select_tool = fig.select(dict(type=BoxSelectTool))
    ## box_select_tool.select_every_mousemove = True
    #tap_tool = fig.select_one(TapTool).renderers = [cr]

    # open a session to keep our local document in sync with server
    #session = push_session(curdoc())

    #curdoc().add_periodic_callback(update, 50)

    def load_file(path):
        image = load_data(path, logger=logger)
        viewer.set_image(image)

    def load_file_cb(attr_name, old_val, new_val):
        #print(attr_name, old_val, new_val)
        load_file(new_val)

    def zoom_ctl_cb(attr_name, old_val, new_val):
        if new_val >= 0:
            new_val += 2
        viewer.zoom_to(int(new_val))
        scale = viewer.get_scale()
        logger.info("%f" % scale)
        viewer.onscreen_message("%f" % (scale), delay=0.3)

    # add a entry widget and configure with the call back
    #dstdir = options.indir
    dstdir = ""
    path_w = TextInput(value=dstdir, title="File:")
    path_w.on_change('value', load_file_cb)

    slide = Slider(start=-20, end=20, step=1, value=1)
    slide.on_change('value', zoom_ctl_cb)

    layout = column(fig, path_w, slide)
    curdoc().add_root(layout)

    if len(args) > 0:
        load_file(args[0])

    # open the document in a browser
    #session.show()

    # run forever
    #session.loop_until_closed()


main(None, sys.argv[1:])

# END



#
# TableView.py -- Table viewer for Ginga
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import logging
from collections import OrderedDict

from ginga.table import AstroTable
from ginga.misc import Callback, Settings
from ginga.misc import Bunch


class TableViewBase(Callback.Callbacks):
    """An abstract base class for displaying tables represented by
    astropy table objects.

    Parameters
    ----------
    logger : :py:class:`~logging.Logger` or `None`
        Logger for tracing and debugging. If not given, one will be created.

    settings : `~ginga.misc.Settings.SettingGroup` or `None`
        Viewer preferences. If not given, one will be created.

    """

    vname = 'Ginga Table'
    vtypes = [AstroTable.AstroTable]

    @classmethod
    def viewable(cls, dataobj):
        """Test whether `dataobj` is viewable by this viewer."""
        if not isinstance(dataobj, AstroTable.AstroTable):
            return False
        return True

    def __init__(self, logger=None, settings=None):
        Callback.Callbacks.__init__(self)

        if logger is not None:
            self.logger = logger
        else:
            self.logger = logging.Logger('TableViewBase')

        # Create settings and set defaults
        if settings is None:
            settings = Settings.SettingGroup(logger=self.logger)
        self.settings = settings

        # for debugging
        self.name = str(self)

        self.settings.add_defaults(color_alternate_rows=True,
                                   max_rows_for_col_resize=5000)

        # For callbacks
        for name in ('table-set', 'configure', ):
            self.enable_callback(name)

    def get_settings(self):
        """Get the settings used by this instance.

        Returns
        -------
        settings : `~ginga.misc.Settings.SettingGroup`
            Settings.

        """
        return self.settings

    def get_logger(self):
        """Get the logger used by this instance.

        Returns
        -------
        logger : :py:class:`~logging.Logger`
            Logger.

        """
        return self.logger

    def set_table(self, table):
        if not isinstance(table, AstroTable.AstroTable):
            raise ValueError("Wrong type of object to load: %s" % (
                str(type(table))))

        self._table = table

        self.make_callback('table-set', table)

    def get_table(self):
        return self._table

    # TO BE DEPRECATED
    get_image = get_table
    set_image = set_table

    # for compatibility with other Ginga viewers
    get_dataobj = get_table
    set_dataobj = set_table

    def initialize_channel(self, fv, channel):
        """The reference viewer calls this method with itself and the channel
        when it is inserted into a channel.
        """
        self.logger.warning("subclass should override this method")


class TableViewGw(TableViewBase):
    """A Ginga viewer for displaying tables of FITS data.
    """

    def __init__(self, logger=None, settings=None):
        super(TableViewGw, self).__init__(logger=logger, settings=settings)

        self.add_callback('table-set', self.set_table_cb)

        from ginga.gw import Widgets

        # Create the viewer as a Treeview widget
        color_alternate = self.settings.get('color_alternate_rows', True)
        self.widget = Widgets.TreeView(auto_expand=True,
                                       sortable=True,
                                       use_alt_row_color=color_alternate)

    def get_widget(self):
        return self.widget

    def set_table_cb(self, viewer, table):
        """Display the given table object."""
        self.clear()
        tree_dict = OrderedDict()

        # Extract data as astropy table
        a_tab = table.get_data()

        # Fill masked values, if applicable
        try:
            a_tab = a_tab.filled()
        except Exception:  # Just use original table
            pass

        # This is to get around table widget not sorting numbers properly
        i_fmt = '{{0:0{0}d}}'.format(len(str(len(a_tab))))

        # Table header with units
        columns = [('Row', '_DISPLAY_ROW')]
        for c in a_tab.columns.values():
            col_str = '{0:^s}\n{1:^s}'.format(c.name, str(c.unit))
            columns.append((col_str, c.name))

        self.widget.setup_table(columns, 1, '_DISPLAY_ROW')

        # Table contents
        for i, row in enumerate(a_tab, 1):
            bnch = Bunch.Bunch(zip(row.colnames, row.as_void()))
            i_str = i_fmt.format(i)
            bnch['_DISPLAY_ROW'] = i_str
            tree_dict[i_str] = bnch

        self.widget.set_tree(tree_dict)

        # Resize column widths
        n_rows = len(tree_dict)
        if n_rows < self.settings.get('max_rows_for_col_resize', 5000):
            self.widget.set_optimal_column_widths()
            self.logger.debug('Resized columns for {0} row(s)'.format(n_rows))

        tablename = table.get('name', 'NoName')
        self.logger.debug('Displayed {0}'.format(tablename))

    def clear(self):
        self.widget.clear()

    def initialize_channel(self, fv, channel):
        # no housekeeping to do (for now) on our part, just override to
        # suppress the logger warning
        pass

# END
